import torch
import torch.nn as nn

from neuralprophet import utils
from neuralprophet.components.future_regressors import FutureRegressors
from neuralprophet.utils_torch import init_parameter

# from neuralprophet.utils_torch import init_parameter


class LinearFutureRegressors(FutureRegressors):
    def __init__(self, config, id_list, quantiles, n_forecasts, device, config_trend_none_bool):
        super().__init__(
            config=config,
            n_forecasts=n_forecasts,
            quantiles=quantiles,
            id_list=id_list,
            device=device,
            config_trend_none_bool=config_trend_none_bool,
        )

        if self.regressors_dims is not None:
            # Regresors params
            self.regressor_params = nn.ParameterDict(
                {
                    # dimensions - [no. of quantiles, no. of additive regressors]
                    "additive": init_parameter(dims=[len(self.quantiles), self.n_additive_regressor_params]),
                    # dimensions - [no. of quantiles, no. of multiplicative regressors]
                    "multiplicative": init_parameter(
                        dims=[len(self.quantiles), self.n_multiplicative_regressor_params]
                    ),
                }
            )

    def scalar_features_effects(self, features, params, indices=None):
        """
        Computes events component of the model

        Parameters
        ----------
            features : torch.Tensor, float
                Features (either additive or multiplicative) related to event component dims (batch, n_forecasts,
                n_features)
            params : nn.Parameter
                Params (either additive or multiplicative) related to events
            indices : list of int
                Indices in the feature tensors related to a particular event
        Returns
        -------
            torch.Tensor
                Forecast component of dims (batch, n_forecasts)
        """
        if indices is not None:
            features = features[:, :, indices]
            params = params[:, indices]

        return torch.sum(features.unsqueeze(dim=2) * params.unsqueeze(dim=0).unsqueeze(dim=0), dim=-1)

    def get_reg_weights(self, name):
        """
        Retrieve the weights of regressor features given the name

        Parameters
        ----------
            name : string
                Regressor name

        Returns
        -------
            torch.tensor
                Weight corresponding to the given regressor
        """

        regressor_dims = self.regressors_dims[name]
        mode = regressor_dims["mode"]
        index = regressor_dims["regressor_index"]

        if mode == "additive":
            regressor_params = self.regressor_params["additive"]
        else:
            assert mode == "multiplicative"
            regressor_params = self.regressor_params["multiplicative"]

        return regressor_params[:, index : (index + 1)]

    def forward(self, inputs, mode, indeces=None):
        """Compute all seasonality components.
        Parameters
        ----------
            f_r : torch.Tensor, float
                future regressors inputs
            mode: string, either "additive" or "multiplicative"
                mode of the regressors
        Returns
        -------
            torch.Tensor
                Forecast component of dims (batch, n_forecasts, no_quantiles)
        """

        if "additive" == mode:
            f_r = self.scalar_features_effects(inputs, self.regressor_params["additive"], indeces)
        if "multiplicative" == mode:
            f_r = self.scalar_features_effects(inputs, self.regressor_params["multiplicative"], indeces)
        return f_r



from abc import abstractmethod

import torch.nn as nn


class BaseComponent(nn.Module):
    def __init__(self, n_forecasts, quantiles, id_list, device):
        super().__init__()
        self.n_forecasts = n_forecasts
        self.quantiles = quantiles
        self.id_list = id_list
        self.device = device

    @abstractmethod
    def forward(self, x):
        """
        Needs to be implemented by subclass.

        Parameters
        ----------
            t : torch.Tensor float
                normalized time, dim: (batch, n_forecasts)
            meta: dict
                Metadata about the all the samples of the model input batch. Contains the following:
                    * ``df_name`` (list, str), time series ID corresponding to each sample of the input batch.
        Returns
        -------
            torch.Tensor
                Component forecast, same dimensions as input t
        """
        pass



import logging
from typing import Optional
from typing import OrderedDict as OrderedDictType
from typing import Tuple

import pandas as pd

from neuralprophet import df_utils
from neuralprophet.configure import ConfigEvents, Regressor
from neuralprophet.data.process import _check_dataframe

log = logging.getLogger("NP.data.splitting")


def _maybe_extend_df(
    df: pd.DataFrame,
    n_forecasts: int,
    max_lags: int,
    freq: Optional[str],
    config_regressors: Optional[OrderedDictType[str, Regressor]],
    config_events: Optional[ConfigEvents],
) -> Tuple[pd.DataFrame, dict]:
    """
    Extend the input DataFrame based on the number of forecasts, maximum lags,
    frequency, regressor configuration, and event configuration.

    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame to be extended.
    n_forecasts : int
        Number of forecasts to be made.
    max_lags : int
        Number of steps ahead of prediction time step to forecast.
    freq : str
        Frequency of the time series data.
    config_regressors : OrderedDict[str, Regressor]
        Configuration of regressors.
    config_events : ConfigEvents
        Configuration of events.

    Returns
    -------
    tuple[pd.DataFrame, int]
        A tuple containing the extended DataFrame and the periods added.
    """
    # Receives df with ID column
    periods_add = {}
    extended_df = pd.DataFrame()
    for df_name, df_i in df.groupby("ID"):
        _ = df_utils.infer_frequency(df_i, n_lags=max_lags, freq=freq)
        # to get all forecasteable values with df given, maybe extend into future:
        periods_add[df_name] = _get_maybe_extend_periods(
            df=df_i, n_forecasts=n_forecasts, max_lags=max_lags, config_regressors=config_regressors
        )
        if periods_add[df_name] > 0:
            # This does not include future regressors or events.
            # periods should be 0 if those are configured.
            last_date = pd.to_datetime(df_i["ds"].copy(deep=True)).sort_values().max()
            future_df = df_utils.make_future_df(
                df_columns=df_i.columns,
                last_date=last_date,
                periods=periods_add[df_name],
                freq=freq,
                config_events=config_events,
            )
            future_df["ID"] = df_name
            df_i = pd.concat([df_i, future_df])
            df_i.reset_index(drop=True, inplace=True)
        extended_df = pd.concat((extended_df, df_i.copy(deep=True)), ignore_index=True)
    return extended_df, periods_add


def _get_maybe_extend_periods(
    df: pd.DataFrame,
    n_forecasts: int,
    max_lags: int,
    config_regressors: Optional[OrderedDictType[str, Regressor]],
) -> int:
    """
    Determine the number of periods to extend the input DataFrame based on the
    number of forecasts, maximum lags, and regressor configuration.


    Parameters
    ----------
    df : pd.DataFrame
        Input DataFrame with a single ID column.
    n_forecasts : int
        Number of steps ahead of prediction time step to forecast.
    max_lags : int
        Maximum number of lags to consider.
    config_regressors : OrderedDictType[str, Regressor]
        Configuration of regressors. If None, the function may extend the
        DataFrame based on `n_forecasts` and `max_lags`.

    Returns
    -------
    int
        Number of periods to extend the input DataFrame.

    Raises
    ------
    AssertionError
        If the input DataFrame contains more than one unique ID.

    Notes
    -----
    The function assumes that the input DataFrame contains columns 'ID' and 'y'.
    """
    # Receives df with single ID column
    assert len(df["ID"].unique()) == 1
    periods_add = 0
    nan_at_end = 0
    while len(df) > nan_at_end and df["y"].isnull().iloc[-(1 + nan_at_end)]:
        nan_at_end += 1
    if max_lags > 0:
        if config_regressors is not None and config_regressors.regressors is not None:
            # if dataframe has already been extended into future,
            # don't extend beyond n_forecasts.
            periods_add = max(0, n_forecasts - nan_at_end)
        else:
            # can not extend as we lack future regressor values.
            periods_add = 0
    return periods_add


def _make_future_dataframe(
    model,
    df: pd.DataFrame,
    events_df: pd.DataFrame,
    regressors_df: pd.DataFrame,
    periods: Optional[int],
    n_historic_predictions: int,
    n_forecasts: int,
    max_lags: int,
    freq: Optional[str],
) -> pd.DataFrame:
    """
    Generate a future dataframe by extending the input dataframe into the future.

    Parameters
    ----------
    model : NeuralProphet
        The model object used for prediction.
    df : pd.DataFrame
        The input dataframe with a single ID column and a 'ds' column containing timestamps.
    events_df : pd.DataFrame, optional
        The dataframe containing information about external events.
    regressors_df : pd.DataFrame, optional
        The dataframe containing information about external regressors.
    periods : int
        The number of steps to extend the DataFrame into the future.
    n_historic_predictions : int
        The number of historic predictions to include in the output dataframe.
    n_forecasts : int
        identical to NeuralProphet
    max_lags : int
        identical to NeuralProphet
    freq : str
        identical to NeuralProphet

    Returns
    -------
    pd.DataFrame
        The extended dataframe with additional rows for future periods.

    Raises
    ------
    ValueError
        If future values of all user specified regressors not provided.
    """
    # Receives df with single ID column
    assert len(df["ID"].unique()) == 1
    if periods == 0 and n_historic_predictions is True:
        log.warning("Not extending df into future as no periods specified." "You can call predict directly instead.")
    df = df.copy(deep=True)
    _ = df_utils.infer_frequency(df, n_lags=max_lags, freq=freq)
    last_date = pd.to_datetime(df["ds"].copy(deep=True).dropna()).sort_values().max()
    if events_df is not None:
        events_df = events_df.copy(deep=True).reset_index(drop=True)
    if regressors_df is not None:
        regressors_df = regressors_df.copy(deep=True).reset_index(drop=True)
    if periods is None:
        periods = 1 if max_lags == 0 else n_forecasts
    else:
        assert periods >= 0

    if isinstance(n_historic_predictions, bool):
        if n_historic_predictions:
            n_historic_predictions = len(df) - max_lags
        else:
            n_historic_predictions = 0
    elif not isinstance(n_historic_predictions, int):
        log.error("non-integer value for n_historic_predictions set to zero.")
        n_historic_predictions = 0

    if periods == 0 and n_historic_predictions == 0:
        raise ValueError("Set either history or future to contain more than zero values.")

    # check for external regressors known in future
    if model.config_regressors.regressors is not None and periods > 0:
        if regressors_df is None:
            raise ValueError("Future values of all user specified regressors not provided")
        else:
            for regressor in model.config_regressors.regressors.keys():
                if regressor not in regressors_df.columns:
                    raise ValueError(f"Future values of user specified regressor {regressor} not provided")

    if len(df) < max_lags:
        raise ValueError(
            "Insufficient input data for a prediction."
            "Please supply historic observations (number of rows) of at least max_lags (max of number of n_lags)."
        )
    elif len(df) < max_lags + n_historic_predictions:
        log.warning(
            f"Insufficient data for {n_historic_predictions} historic forecasts, reduced to {len(df) - max_lags}."
        )
        n_historic_predictions = len(df) - max_lags
    if (n_historic_predictions + max_lags) == 0:
        df = pd.DataFrame(columns=df.columns)
    else:
        df = df[-(max_lags + n_historic_predictions) :]
        nan_at_end = 0
        while len(df) > nan_at_end and df["y"].isnull().iloc[-(1 + nan_at_end)]:
            nan_at_end += 1
        if nan_at_end > 0:
            if max_lags > 0 and (nan_at_end + 1) >= max_lags:
                raise ValueError(
                    f"{nan_at_end + 1} missing values were detected at the end of df before df was extended into "
                    "the future. Please make sure there are no NaN values at the end of df."
                )
            df["y"].iloc[-(nan_at_end + 1) :].ffill(inplace=True)
            log.warning(
                f"{nan_at_end + 1} missing values were forward-filled at the end of df before df was extended into the "
                "future. Please make sure there are no NaN values at the end of df."
            )

    if len(df) > 0:
        if len(df.columns) == 1 and "ds" in df:
            assert max_lags == 0
            df = _check_dataframe(model, df, check_y=False, exogenous=False)
        else:
            df = _check_dataframe(model, df, check_y=max_lags > 0, exogenous=True, future=True)
    # future data
    # check for external events known in future
    if model.config_events is not None and periods > 0 and events_df is None:
        log.warning(
            "Future values not supplied for user specified events. "
            "All events being treated as not occurring in future"
        )

    if max_lags > 0:
        if periods > 0 and periods != n_forecasts:
            periods = n_forecasts
            log.warning(f"Number of forecast steps is defined by n_forecasts. Adjusted to {n_forecasts}.")

    if periods > 0:
        future_df = df_utils.make_future_df(
            df_columns=df.columns,
            last_date=last_date,
            periods=periods,
            freq=freq,
            config_events=model.config_events,
            events_df=events_df,
            config_regressors=model.config_regressors,
            regressors_df=regressors_df,
        )
        if len(df) > 0:
            df = pd.concat([df, future_df])
        else:
            df = future_df
    df = df.reset_index(drop=True)
    model.predict_steps = periods
    return df



import inspect
import logging
from typing import Any

import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn as nn
from captum.attr import Saliency

log = logging.getLogger("NP.utils_torch")


def init_parameter(dims):
    """
    Create and initialize a new torch Parameter.

    Parameters
    ----------
        dims : list or tuple
            Desired dimensions of parameter

    Returns
    -------
        nn.Parameter
            initialized Parameter
    """
    if len(dims) > 1:
        return nn.Parameter(nn.init.xavier_normal_(torch.randn(dims)), requires_grad=True)
    else:
        return nn.Parameter(torch.nn.init.xavier_normal_(torch.randn([1] + dims)).squeeze(0), requires_grad=True)


def penalize_nonzero(weights, eagerness=1.0, acceptance=1.0):
    cliff = 1.0 / (np.e * eagerness)
    return torch.log(cliff + acceptance * torch.abs(weights)) - np.log(cliff)


def create_optimizer_from_config(optimizer_name, optimizer_args):
    """
    Translate the optimizer name and arguments into a torch optimizer.
    If an optimizer object is provided, it is returned as is.
    The optimizer is not initialized yet since this is done by the trainer.

    Parameters
        ----------
            optimizer_name : int
                Object provided to NeuralProphet as optimizer.
            optimizer_args : dict
                Arguments for the optimizer.

        Returns
        -------
            optimizer : torch.optim.Optimizer
                The optimizer object.
            optimizer_args : dict
                The optimizer arguments.
    """
    if isinstance(optimizer_name, str):
        if optimizer_name.lower() == "adamw":
            # Tends to overfit, but reliable
            optimizer = torch.optim.AdamW
            optimizer_args["weight_decay"] = 1e-3
        elif optimizer_name.lower() == "sgd":
            # better validation performance, but diverges sometimes
            optimizer = torch.optim.SGD
            optimizer_args["momentum"] = 0.9
            optimizer_args["weight_decay"] = 1e-4
        else:
            raise ValueError(f"The optimizer name {optimizer_name} is not supported.")
    elif inspect.isclass(optimizer_name) and issubclass(optimizer_name, torch.optim.Optimizer):
        optimizer = optimizer_name
    else:
        raise ValueError("The provided optimizer is not supported.")
    return optimizer, optimizer_args


def interprete_model(
    target_model: pl.LightningModule,
    net: str,
    forward_func: str,
    _num_in_features: int = None,
    _num_out_features: int = None,
    _input: torch.Tensor = None,
    additional_forward_args: Any = None,
):
    """
    Returns model input attributions for a given network and forward function.

    Parameters
    ----------
        target_model : pl.LightningModule
            The model for which input attributions are to be computed.
        net : str
            Name of the network for which input attributions are to be computed.
        forward_func : str
            Name of the forward function for which input attributions are to be computed.

        _input : torch.Tensor
            Input for which the attributions are to be computed.

    Returns
    -------
        torch.Tensor
            Input attributions for the given network and forward function.
    """
    # Load the respective forward function from the model and init model interpreter
    forward = getattr(target_model, forward_func)
    saliency = Saliency(forward_func=forward)

    # Number of quantiles
    num_quantiles = len(target_model.quantiles)

    # Number of input features to the net (aka n_lags)
    num_in_features = getattr(target_model, net)[0].in_features if _num_in_features is None else _num_in_features
    # Number of output features from the net (aka n_forecasts)
    num_out_features = getattr(target_model, net)[-1].out_features if _num_out_features is None else _num_out_features

    num_out_features_without_quantiles = int(num_out_features / num_quantiles)

    # Create a tensor of ones as model input
    model_input = torch.ones(1, num_in_features, requires_grad=True) if _input is None else _input

    # Iterate through each output feature and compute the model attribution wrt. the input
    attributions = torch.empty((0, num_in_features))
    for output_feature in range(num_out_features_without_quantiles):
        for quantile in range(num_quantiles):
            target_attribution = saliency.attribute(
                model_input,
                target=[(output_feature, quantile)],
                abs=False,
                additional_forward_args=additional_forward_args,
            )
            attributions = torch.cat((attributions, target_attribution), 0)

    # Average the attributions over the input features
    # Idea: Average attribution of each lag on all forecasts (eg the n'th lag has an attribution of xyz on the forecast)
    # TODO: support the visualization of 2d tensors in plot_parameters
    # (aka the attribution of the n'th lag on the m'th forecast)

    return attributions



import logging

import torchmetrics

log = logging.getLogger("NP.metrics")

METRICS = {
    # "short_name": [torchmetrics.Metric name, {optional args}]
    "MAE": ["MeanAbsoluteError", {}],
    "MSE": ["MeanSquaredError", {"squared": True}],
    "RMSE": ["MeanSquaredError", {"squared": False}],
}


def get_metrics(metric_input):
    """
    Returns a dict of metrics.

    Parameters
    ----------
        metrics : input received from the user
            List of metrics to use.

    Returns
    -------
        dict
            Dict of names of torchmetrics.Metric metrics
    """
    if metric_input is None:
        return {}
    elif metric_input is True:
        return {"MAE": METRICS["MAE"], "RMSE": METRICS["RMSE"]}
    elif isinstance(metric_input, str):
        if metric_input.upper() in METRICS.keys():
            return {metric_input: METRICS[metric_input.upper()]}
        else:
            raise ValueError("Received unsupported argument for collect_metrics.")
    elif isinstance(metric_input, list):
        if all([m.upper() in METRICS.keys() for m in metric_input]):
            return {m: METRICS[m.upper()] for m in metric_input}
        else:
            raise ValueError("Received unsupported argument for collect_metrics.")
    elif isinstance(metric_input, dict):
        # check if all values are names belonging to torchmetrics.Metric
        try:
            for _metric in metric_input.values():
                torchmetrics.__dict__[_metric]()
        except KeyError:
            raise ValueError(
                "Received unsupported argument for collect_metrics."
                "All metrics must be valid names of torchmetrics.Metric objects."
            )
        return {k: [v, {}] for k, v in metric_input.items()}
    elif metric_input is not False:
        raise ValueError("Received unsupported argument for collect_metrics.")



import logging
from collections import OrderedDict, defaultdict
from datetime import datetime
from typing import Optional

import numpy as np
import pandas as pd
import torch
from torch.utils.data.dataset import Dataset

from neuralprophet import configure, utils
from neuralprophet.df_utils import get_max_num_lags
from neuralprophet.hdays_utils import get_country_holidays

log = logging.getLogger("NP.time_dataset")


class GlobalTimeDataset(Dataset):
    def __init__(self, df, **kwargs):
        """Initialize Timedataset from time-series df.
        Parameters
        ----------
            df : pd.DataFrame
                dataframe containing column ``ds``, ``y``, and optionally``ID`` and
                normalized columns normalized columns ``ds``, ``y``, ``t``, ``y_scaled``
            **kwargs : dict
                Identical to :meth:`tabularize_univariate_datetime`
        """
        # # TODO (future): vectorize
        timedatasets = [TimeDataset(df_i, df_name, **kwargs) for df_name, df_i in df.groupby("ID")]
        self.combined_timedataset = [item for timedataset in timedatasets for item in timedataset]
        self.length = sum(timedataset.length for timedataset in timedatasets)

    def __len__(self):
        return self.length

    def __getitem__(self, idx):
        return self.combined_timedataset[idx]


class TimeDataset(Dataset):
    """Create a PyTorch dataset of a tabularized time-series"""

    def __init__(self, df, name, **kwargs):
        """Initialize Timedataset from time-series df.
        Parameters
        ----------
            df : pd.DataFrame
                Time series data
            name : str
                Name of time-series
            **kwargs : dict
                Identical to :meth:`tabularize_univariate_datetime`
        """
        self.name = name
        self.length = None
        self.inputs = OrderedDict({})
        self.targets = None
        self.meta = OrderedDict({})
        self.two_level_inputs = [
            "seasonalities",
            "covariates",
            "events",
            "regressors",
        ]
        inputs, targets, drop_missing = tabularize_univariate_datetime(df, **kwargs)
        self.init_after_tabularized(inputs, targets)
        self.filter_samples_after_init(kwargs["prediction_frequency"])
        self.drop_nan_after_init(df, kwargs["predict_steps"], drop_missing)

    def drop_nan_after_init(self, df, predict_steps, drop_missing):
        """Checks if inputs/targets contain any NaN values and drops them, if user opts to.
        Parameters
        ----------
            drop_missing : bool
                whether to automatically drop missing samples from the data
            predict_steps : int
                number of steps to predict
        """
        nan_idx = []
        # NaNs in inputs
        for key, data in self.inputs.items():
            if isinstance(data, torch.Tensor):
                nans = torch.where(torch.isnan(data))[0].tolist()
                if len(nans) > 0:
                    nan_idx += nans
            elif isinstance(data, dict):
                for subkey, subdata in data.items():
                    nans = torch.where(torch.isnan(subdata))[0].tolist()
                    if len(nans) > 0:
                        nan_idx += nans

        # NaNs in targets that are not inserted for prediction at the end
        nans = torch.where(torch.isnan(self.targets))[0].tolist()
        if len(nans) > 0:
            for idx in nans:
                if idx not in nan_idx and idx < len(self) - predict_steps:
                    nan_idx.append(idx)

        nan_idx = list(set(nan_idx))
        nan_idx.sort()
        if drop_missing and len(nan_idx) > 0:
            log.warning(f"{len(nan_idx)} samples with missing values were dropped from the data. ")
            for key, data in self.inputs.items():
                if key not in ["time", "lags"]:  # "time_lagged"
                    for name, features in data.items():
                        self.inputs[key][name] = np.delete(self.inputs[key][name], nan_idx, 0)
                else:
                    self.inputs[key] = np.delete(self.inputs[key], nan_idx, 0)
            self.targets = np.delete(self.targets, nan_idx, 0)
            self.length = self.inputs["time"].shape[0]
        if not drop_missing and len(nan_idx) > 0:
            raise ValueError(
                "Inputs/targets with missing values detected. "
                "Please either adjust imputation parameters, or set 'drop_missing' to True to drop those samples."
            )

    @staticmethod
    def _split_nested_dict(inputs):
        """Split nested dict into list of dicts.
        Parameters
        ----------
            inputs : ordered dict
                Nested dict to be split.
        Returns
        -------
            list of dicts
                List of dicts with same keys as inputs.
        """

        def split_dict(inputs, index):
            return {k: v[index] if not isinstance(v, dict) else split_dict(v, index) for k, v in inputs.items()}

        length = next(iter(inputs.values())).shape[0]
        return [split_dict(inputs, i) for i in range(length)]

    def init_after_tabularized(self, inputs, targets=None):
        """Create Timedataset with data.
        Parameters
        ----------
            inputs : ordered dict
                Identical to returns from :meth:`tabularize_univariate_datetime`
            targets : np.array, float
                Identical to returns from :meth:`tabularize_univariate_datetime`
        """
        inputs_dtype = {
            "time": torch.float,
            "timestamps": np.datetime64,
            "seasonalities": torch.float,
            "events": torch.float,
            "lags": torch.float,
            "covariates": torch.float,
            "regressors": torch.float,
        }
        targets_dtype = torch.float
        self.length = inputs["time"].shape[0]

        for key, data in inputs.items():
            if key in self.two_level_inputs:
                self.inputs[key] = OrderedDict({})
                for name, features in data.items():
                    if features.dtype != np.float32:
                        features = features.astype(np.float32, copy=False)

                    tensor = torch.from_numpy(features)

                    if tensor.dtype != inputs_dtype[key]:
                        self.inputs[key][name] = tensor.to(
                            dtype=inputs_dtype[key]
                        )  # this can probably be removed, but was included in the previous code
                    else:
                        self.inputs[key][name] = tensor
            else:
                if key == "timestamps":
                    self.inputs[key] = data
                else:
                    self.inputs[key] = torch.from_numpy(data).type(inputs_dtype[key])
        self.targets = torch.from_numpy(targets).type(targets_dtype).unsqueeze(dim=2)
        self.meta["df_name"] = self.name
        self.samples = self._split_nested_dict(self.inputs)

    def filter_samples_after_init(
        self,
        prediction_frequency=None,
    ):
        """Filters samples from the dataset based on the forecast frequency.
        Parameters
        ----------
            prediction_frequency : int
                periodic interval in which forecasts should be made.
            Note
            ----
            E.g. if prediction_frequency=7, forecasts are only made on every 7th step (once in a week in case of daily
            resolution).
        """
        if prediction_frequency is None or prediction_frequency == 1:
            return
        # Only the first target timestamp is of interest for filtering
        timestamps = pd.to_datetime([sample["timestamps"][0] for sample in self.samples])
        masks = []
        for key, value in prediction_frequency.items():
            if key == "daily-hour":
                mask = timestamps.hour == value + 1  # because prediction starts one step after origin
            elif key == "weekly-day":
                mask = timestamps.dayofweek == value + 1
            elif key == "monthly-day":
                mask = timestamps.day == value + 1
            elif key == "yearly-month":
                mask = timestamps.month == value + 1
            elif key == "hourly-minute":
                mask = timestamps.minute == value + 1
            else:
                raise ValueError(f"Invalid prediction frequency: {key}")
            masks.append(mask)
        mask = np.ones((len(timestamps),), dtype=bool)
        for m in masks:
            mask = mask & m
        self.samples = [self.samples[i] for i in range(len(self.samples)) if mask[i]]

        # Exact timestamps are not needed anymore
        self.inputs.pop("timestamps")
        for sample in self.samples:
            sample.pop("timestamps")
        self.length = len(self.samples)

    def __getitem__(self, index):
        """Overrides parent class method to get an item at index.
        Parameters
        ----------
            index : int
                Sample location in dataset
        Returns
        -------
        OrderedDict
            Model inputs, each of len(df) but with varying dimensions
            Note
            ----
            Contains the following data:
            Model Inputs
                * ``time`` (np.array, float), dims: (num_samples, 1)
                * ``seasonalities`` (OrderedDict), named seasonalities
                each with features (np.array, float) - dims: (num_samples, n_features[name])
                * ``lags`` (np.array, float), dims: (num_samples, n_lags)
                * ``covariates`` (OrderedDict), named covariates,
                each with features (np.array, float) of dims: (num_samples, n_lags)
                * ``events`` (OrderedDict), events,
                each with features (np.array, float) of dims: (num_samples, n_lags)
                * ``regressors`` (OrderedDict), regressors,
                each with features (np.array, float) of dims: (num_samples, n_lags)
        np.array, float
            Targets to be predicted of same length as each of the model inputs, dims: (num_samples, n_forecasts)
        """
        sample = self.samples[index]
        targets = self.targets[index]
        meta = self.meta
        return sample, targets, meta

    def __len__(self):
        """Overrides Parent class method to get data length."""
        return self.length


def tabularize_univariate_datetime(
    df,
    predict_mode=False,
    n_lags=0,
    n_forecasts=1,
    predict_steps=1,
    config_seasonality: Optional[configure.ConfigSeasonality] = None,
    config_events: Optional[configure.ConfigEvents] = None,
    config_country_holidays=None,
    config_lagged_regressors: Optional[configure.ConfigLaggedRegressors] = None,
    config_regressors: Optional[configure.ConfigFutureRegressors] = None,
    config_missing=None,
    prediction_frequency=None,
):
    """Create a tabular dataset from univariate timeseries for supervised forecasting.
    Note
    ----
    Data must have no gaps.
    If data contains missing values, they are ignored for the creation of the dataset.
    Parameters
    ----------
        df : pd.DataFrame
            Sequence of observations with original ``ds``, ``y`` and normalized ``t``, ``y_scaled`` columns
        config_seasonality : configure.ConfigSeasonality
            Configuration for seasonalities
        n_lags : int
            Number of lagged values of series to include as model inputs (aka AR-order)
        n_forecasts : int
            Number of steps to forecast into future
        config_events : configure.ConfigEvents
            User specified events, each with their upper, lower windows (int) and regularization
        config_country_holidays : configure.ConfigCountryHolidays
            Configurations (holiday_names, upper, lower windows, regularization) for country specific holidays
        config_lagged_regressors : configure.ConfigLaggedRegressors
            Configurations for lagged regressors
        config_regressors : configure.ConfigFutureRegressors
            Configuration for regressors
        predict_mode : bool
            Chooses the prediction mode
            Options
                * (default) ``False``: Includes target values
                * ``True``: Does not include targets but includes entire dataset as input
    Returns
    -------
        OrderedDict
            Model inputs, each of len(df) but with varying dimensions
            Note
            ----
            Contains the following data:
            Model Inputs
                * ``time`` (np.array, float), dims: (num_samples, 1)
                * ``seasonalities`` (OrderedDict), named seasonalities
                each with features (np.array, float) - dims: (num_samples, n_features[name])
                * ``lags`` (np.array, float), dims: (num_samples, n_lags)
                * ``covariates`` (OrderedDict), named covariates,
                each with features (np.array, float) of dims: (num_samples, n_lags)
                * ``events`` (OrderedDict), events,
                each with features (np.array, float) of dims: (num_samples, n_lags)
                * ``regressors`` (OrderedDict), regressors,
                each with features (np.array, float) of dims: (num_samples, n_lags)
        np.array, float
            Targets to be predicted of same length as each of the model inputs, dims: (num_samples, n_forecasts)
    """
    max_lags = get_max_num_lags(config_lagged_regressors, n_lags)
    n_samples = len(df) - max_lags + 1 - n_forecasts
    # data is stored in OrderedDict
    inputs = OrderedDict({})

    def _stride_time_features_for_forecasts(x):
        window_size = n_lags + n_forecasts

        if x.ndim == 1:
            shape = (n_samples, window_size)
        else:
            shape = (n_samples, window_size) + x.shape[1:]

        stride = x.strides[0]
        strides = (stride, stride) + x.strides[1:]
        start_index = max_lags - n_lags
        return np.lib.stride_tricks.as_strided(x[start_index:], shape=shape, strides=strides)

    def _stride_future_time_features_for_forecasts(x):
        return np.array([x[max_lags + i : max_lags + i + n_forecasts] for i in range(n_samples)], dtype=x.dtype)

    def _stride_lagged_features(df_col_name, feature_dims):
        # only for case where max_lags > 0
        assert feature_dims >= 1
        series = df.loc[:, df_col_name].values
        # Added dtype=np.float64 to solve the problem with np.isnan for ubuntu test
        return np.array(
            [series[i + max_lags - feature_dims : i + max_lags] for i in range(n_samples)], dtype=np.float32
        )

    def _stride_timestamps_for_forecasts(x):
        # only for case where n_lags > 0
        if x.dtype != np.float64:
            dtype = np.datetime64
        else:
            dtype = np.float64
        return np.array([x[i + max_lags : i + max_lags + n_forecasts] for i in range(n_samples)], dtype=dtype)

    # time is the time at each forecast step
    t = df.loc[:, "t"].values
    if max_lags == 0:
        assert n_forecasts == 1
        time = np.expand_dims(t, 1)
    else:
        time = _stride_time_features_for_forecasts(t)
    inputs["time"] = time  # contains n_lags + n_forecasts

    if prediction_frequency is not None:
        ds = df.loc[:, "ds"].values
        if max_lags == 0:  # is it rather n_lags?
            timestamps = np.expand_dims(ds, 1)
        else:
            timestamps = _stride_timestamps_for_forecasts(ds)
        inputs["timestamps"] = timestamps

    if config_seasonality is not None:
        seasonalities = seasonal_features_from_dates(df, config_seasonality)
        for name, features in seasonalities.items():
            if max_lags == 0:
                seasonalities[name] = np.expand_dims(features, axis=1)
            else:
                # stride into num_forecast at dim=1 for each sample, just like we did with time
                seasonalities[name] = _stride_time_features_for_forecasts(features)
        inputs["seasonalities"] = seasonalities

    if n_lags > 0 and "y" in df.columns:
        inputs["lags"] = _stride_lagged_features(df_col_name="y_scaled", feature_dims=n_lags)

    if config_lagged_regressors is not None and max_lags > 0:
        covariates = OrderedDict({})
        for covar in df.columns:
            if covar in config_lagged_regressors:
                assert config_lagged_regressors[covar].n_lags > 0
                window = config_lagged_regressors[covar].n_lags
                covariates[covar] = _stride_lagged_features(df_col_name=covar, feature_dims=window)
        inputs["covariates"] = covariates

    # get the regressors features
    if config_regressors is not None and config_regressors.regressors is not None:
        additive_regressors, multiplicative_regressors = make_regressors_features(df, config_regressors)

        regressors = OrderedDict({})
        if max_lags == 0:
            if additive_regressors is not None:
                regressors["additive"] = np.expand_dims(additive_regressors, axis=1)
            if multiplicative_regressors is not None:
                regressors["multiplicative"] = np.expand_dims(multiplicative_regressors, axis=1)
        else:
            if additive_regressors is not None:
                additive_regressor_feature_windows = []
                # additive_regressor_feature_windows_lagged = []
                for i in range(0, additive_regressors.shape[1]):
                    # stride into num_forecast at dim=1 for each sample, just like we did with time
                    stride = _stride_time_features_for_forecasts(additive_regressors[:, i])
                    additive_regressor_feature_windows.append(stride)
                additive_regressors = np.dstack(additive_regressor_feature_windows)
                regressors["additive"] = additive_regressors

            if multiplicative_regressors is not None:
                multiplicative_regressor_feature_windows = []
                for i in range(0, multiplicative_regressors.shape[1]):
                    stride = _stride_time_features_for_forecasts(multiplicative_regressors[:, i])
                    multiplicative_regressor_feature_windows.append(stride)
                multiplicative_regressors = np.dstack(multiplicative_regressor_feature_windows)
                regressors["multiplicative"] = multiplicative_regressors
        inputs["regressors"] = regressors

    # get the events features
    if config_events is not None or config_country_holidays is not None:
        additive_events, multiplicative_events = make_events_features(df, config_events, config_country_holidays)

        events = OrderedDict({})
        if max_lags == 0:
            if additive_events is not None:
                events["additive"] = np.expand_dims(additive_events, axis=1)
            if multiplicative_events is not None:
                events["multiplicative"] = np.expand_dims(multiplicative_events, axis=1)
        else:
            if additive_events is not None:
                additive_event_feature_windows = []
                for i in range(0, additive_events.shape[1]):
                    # stride into num_forecast at dim=1 for each sample, just like we did with time
                    additive_event_feature_windows.append(_stride_time_features_for_forecasts(additive_events[:, i]))
                additive_events = np.dstack(additive_event_feature_windows)
                events["additive"] = additive_events

            if multiplicative_events is not None:
                multiplicative_event_feature_windows = []
                # multiplicative_event_feature_windows_lagged = []
                for i in range(0, multiplicative_events.shape[1]):
                    # stride into num_forecast at dim=1 for each sample, just like we did with time
                    multiplicative_event_feature_windows.append(
                        _stride_time_features_for_forecasts(multiplicative_events[:, i])
                    )
                multiplicative_events = np.dstack(multiplicative_event_feature_windows)
                events["multiplicative"] = multiplicative_events
        inputs["events"] = events

    if predict_mode:
        targets = np.empty_like(time[:, n_lags:])
        targets = np.nan_to_num(targets)
    else:
        targets = _stride_future_time_features_for_forecasts(df["y_scaled"].values)

    tabularized_input_shapes_str = ""
    for key, value in inputs.items():
        if key in [
            "seasonalities",
            "covariates",
            "events",
            "regressors",
        ]:
            for name, period_features in value.items():
                tabularized_input_shapes_str += f"    {name} {key} {period_features}\n"
        else:
            tabularized_input_shapes_str += f"    {key} {value.shape} \n"
    log.debug(f"Tabularized inputs shapes: \n{tabularized_input_shapes_str}")

    return inputs, targets, config_missing.drop_missing


def fourier_series(dates, period, series_order):
    """Provides Fourier series components with the specified frequency and order.
    Note
    ----
    Identical to OG Prophet.
    Parameters
    ----------
        dates : pd.Series
            Containing timestamps
        period : float
            Number of days of the period
        series_order : int
            Number of fourier components
    Returns
    -------
        np.array
            Matrix with seasonality features
    """
    # convert to days since epoch
    t = np.array((dates - datetime(1970, 1, 1)).dt.total_seconds().astype(np.float32)) / (3600 * 24.0)
    return fourier_series_t(t, period, series_order)


def fourier_series_t(t, period, series_order):
    """Provides Fourier series components with the specified frequency and order.
    Note
    ----
    This function is identical to Meta AI's Prophet Library
    Parameters
    ----------
        t : pd.Series, float
            Containing time as floating point number of days
        period : float
            Number of days of the period
        series_order : int
            Number of fourier components
    Returns
    -------
        np.array
            Matrix with seasonality features
    """
    features = np.column_stack(
        [fun((2.0 * (i + 1) * np.pi * t / period)) for i in range(series_order) for fun in (np.sin, np.cos)]
    )
    return features


def make_country_specific_holidays_df(year_list, country):
    """
    Make dataframe of country specific holidays for given years and countries
    Parameters
    ----------
        year_list : list
            List of years
        country : str, list
            List of country names
    Returns
    -------
        pd.DataFrame
            Containing country specific holidays df with columns 'ds' and 'holiday'
    """
    # iterate over countries and get holidays for each country
    # convert to list if not already
    if isinstance(country, str):
        country = [country]
    country_specific_holidays = {}
    for single_country in country:
        single_country_specific_holidays = get_country_holidays(single_country, year_list)
        # only add holiday if it is not already in the dict
        country_specific_holidays.update(single_country_specific_holidays)
    country_specific_holidays_dict = defaultdict(list)
    for date, holiday in country_specific_holidays.items():
        country_specific_holidays_dict[holiday].append(pd.to_datetime(date))
    return country_specific_holidays_dict


def _create_event_offset_features(event, config, feature, additive_events, multiplicative_events):
    """
    Create event offset features for the given event, config and feature
    Parameters
    ----------
        event : str
            Name of the event
        config : configure.ConfigEvents
            User specified events, holidays, and country specific holidays
        feature : pd.Series
            Feature for the event
        additive_events : pd.DataFrame
            Dataframe of additive events
        multiplicative_events : pd.DataFrame
            Dataframe of multiplicative events
    Returns
    -------
        tuple
            Tuple of additive_events and multiplicative_events
    """
    lw = config.lower_window
    uw = config.upper_window
    mode = config.mode
    for offset in range(lw, uw + 1):
        key = utils.create_event_names_for_offsets(event, offset)
        offset_feature = feature.shift(periods=offset, fill_value=0.0)
        if mode == "additive":
            additive_events[key] = offset_feature
        else:
            multiplicative_events[key] = offset_feature


def make_events_features(df, config_events: Optional[configure.ConfigEvents] = None, config_country_holidays=None):
    """
    Construct arrays of all event features
    Parameters
    ----------
        df : pd.DataFrame
            Dataframe with all values including the user specified events (provided by user)
        config_events : configure.ConfigEvents
            User specified events, each with their upper, lower windows (int), regularization
        config_country_holidays : configure.ConfigCountryHolidays
            Configurations (holiday_names, upper, lower windows, regularization) for country specific holidays
    Returns
    -------
        np.array
            All additive event features (both user specified and country specific)
        np.array
            All multiplicative event features (both user specified and country specific)
    """
    df = df.reset_index(drop=True)
    additive_events = pd.DataFrame()
    multiplicative_events = pd.DataFrame()

    # create all user specified events
    if config_events is not None:
        for event, configs in config_events.items():
            feature = df[event]
            _create_event_offset_features(event, configs, feature, additive_events, multiplicative_events)

    # create all country specific holidays
    if config_country_holidays is not None:
        year_list = list({x.year for x in df.ds})
        country_holidays_dict = make_country_specific_holidays_df(year_list, config_country_holidays.country)
        for holiday in config_country_holidays.holiday_names:
            feature = pd.Series([0.0] * df.shape[0])
            if holiday in country_holidays_dict.keys():
                dates = country_holidays_dict[holiday]
                feature[df.ds.isin(dates)] = 1.0
            _create_event_offset_features(
                holiday, config_country_holidays, feature, additive_events, multiplicative_events
            )

    # Make sure column order is consistent
    if not additive_events.empty:
        additive_events = additive_events[sorted(additive_events.columns.tolist())]
        additive_events = additive_events.values
    else:
        additive_events = None
    if not multiplicative_events.empty:
        multiplicative_events = multiplicative_events[sorted(multiplicative_events.columns.tolist())]
        multiplicative_events = multiplicative_events.values
    else:
        multiplicative_events = None

    return additive_events, multiplicative_events


def make_regressors_features(df, config_regressors):
    """Construct arrays of all scalar regressor features
    Parameters
    ----------
        df : pd.DataFrame
            Dataframe with all values including the user specified regressors
        config_regressors : configure.ConfigFutureRegressors
            User specified regressors config
    Returns
    -------
        np.array
            All additive regressor features
        np.array
            All multiplicative regressor features
    """
    additive_regressors = pd.DataFrame()
    multiplicative_regressors = pd.DataFrame()

    for reg in df.columns:
        if reg in config_regressors.regressors:
            mode = config_regressors.regressors[reg].mode
            if mode == "additive":
                additive_regressors[reg] = df[reg]
            else:
                multiplicative_regressors[reg] = df[reg]

    if not additive_regressors.empty:
        additive_regressors = additive_regressors[sorted(additive_regressors.columns.tolist())]
        additive_regressors = additive_regressors.values
    else:
        additive_regressors = None
    if not multiplicative_regressors.empty:
        multiplicative_regressors = multiplicative_regressors[sorted(multiplicative_regressors.columns.tolist())]
        multiplicative_regressors = multiplicative_regressors.values
    else:
        multiplicative_regressors = None

    return additive_regressors, multiplicative_regressors


def seasonal_features_from_dates(df, config_seasonality: configure.ConfigSeasonality):
    """Dataframe with seasonality features.
    Includes seasonality features, holiday features, and added regressors.
    Parameters
    ----------
        df : pd.DataFrame
            Dataframe with all values
        config_seasonality : configure.ConfigSeasonality
            Configuration for seasonalities
    Returns
    -------
        OrderedDict
            Dictionary with keys for each period name containing an np.array
            with the respective regression features. each with dims: (len(dates), 2*fourier_order)
    """
    dates = df["ds"]
    assert len(dates.shape) == 1
    seasonalities = OrderedDict({})
    # Seasonality features
    for name, period in config_seasonality.periods.items():
        if period.resolution > 0:
            if config_seasonality.computation == "fourier":
                features = fourier_series(
                    dates=dates,
                    period=period.period,
                    series_order=period.resolution,
                )
            else:
                raise NotImplementedError
            if period.condition_name is not None:
                features = features * df[period.condition_name].values[:, np.newaxis]
            seasonalities[name] = features
    return seasonalities



from typing import Iterable, Optional, Union

import holidays


def get_country_holidays(country: str, years: Optional[Union[int, Iterable[int]]] = None):
    """
    Helper function to get holidays for a country.

    Parameters
    ----------
        country : str
            Country name to retrieve country specific holidays
        years : int, list
            Year or list of years to retrieve holidays for

    Returns
    -------
        set
            All possible holiday dates and names of given country

    """
    substitutions = {
        "TU": "TR",  # For compatibility with Turkey as "TU" cases.
    }

    country = substitutions.get(country, country)
    if not hasattr(holidays, country):
        raise AttributeError(f"Holidays in {country} are not currently supported!")

    return getattr(holidays, country)(years=years)



import logging

import numpy as np

from neuralprophet.plot_model_parameters_matplotlib import plot_custom_season, plot_daily, plot_weekly, plot_yearly
from neuralprophet.plot_utils import set_y_as_percent

log = logging.getLogger("NP.plotting")

try:
    from matplotlib import pyplot as plt
    from matplotlib.dates import AutoDateFormatter, AutoDateLocator
    from pandas.plotting import deregister_matplotlib_converters

    deregister_matplotlib_converters()
except ImportError:
    log.error("Importing matplotlib failed. Plotting will not work.")


def plot(
    fcst,
    quantiles,
    ax=None,
    xlabel="ds",
    ylabel="y",
    highlight_forecast=None,
    line_per_origin=False,
    figsize=(10, 6),
):
    """Plot the NeuralProphet forecast.

    Parameters
    ----------
        fcst : pd.DataFrame
            Output of m.predict
        quantiles: list
            Quantiles for which the forecasts are to be plotted
        ax : matplotlib axes
            Axes to plot on
        xlabel : str
            Label name on X-axis
        ylabel : str
            Label name on Y-axis
        highlight_forecast : int
            i-th step ahead forecast to highlight.
        line_per_origin : bool
            Print a line per forecast of one per forecast age
        figsize : tuple
            Width, height in inches.

    Returns
    -------
        matplotlib.pyplot.figure
            Figure showing the NeuralProphet forecast

    Examples
    --------
        Base usage

        >>> from neuralprophet import NeuralProphet
        >>> m = NeuralProphet()
        >>> metrics = m.fit(df, freq="D")
        >>> future = m.make_future_dataframe(df=df, periods=365)
        >>> forecast = m.predict(df=future)
        >>> fig_forecast = m.plot(forecast)

        Additional plot specifications

        >>> m.plot(forecast,
        >>>    xlabel="ds",
        >>>    ylabel="y",
        >>>    highlight_forecast=None,
        >>>    line_per_origin=False,
        >>>    figsize=(10, 6)
        >>>    )

    """
    fcst = fcst.fillna(value=np.nan)
    if ax is None:
        fig = plt.figure(facecolor="w", figsize=figsize)
        ax = fig.add_subplot(111)
    else:
        fig = ax.get_figure()
    ds = fcst["ds"].dt.to_pydatetime()
    colname = "yhat"
    step = 1
    # if plot_latest_forecast(), column names become "origin-x", with origin-0 being the latest forecast
    if line_per_origin:
        colname = "origin-"
        step = 0
    # all yhat column names without quantiles
    yhat_col_names = [col_name for col_name in fcst.columns if col_name.startswith(colname) and "%" not in col_name]

    if highlight_forecast is None or line_per_origin:
        for i, name in enumerate(yhat_col_names):
            ax.plot(
                ds,
                fcst[f"{colname}{i if line_per_origin else i + 1}"],
                ls="-",
                c="#0072B2",
                alpha=0.2 + 2.0 / (i + 2.5),
                label=name,
            )

        if len(quantiles) > 1:
            for i in range(1, len(quantiles)):
                ax.fill_between(
                    ds,
                    fcst[f"{colname}{step}"],
                    fcst[f"{colname}{step} {round(quantiles[i] * 100, 1)}%"],
                    color="#0072B2",
                    alpha=0.2,
                )

    if highlight_forecast is not None:
        if line_per_origin:
            num_forecast_steps = sum(fcst["origin-0"].notna())
            steps_from_last = num_forecast_steps - highlight_forecast
            for i in range(len(yhat_col_names)):
                x = ds[-(1 + i + steps_from_last)]
                y = fcst[f"origin-{i}"].values[-(1 + i + steps_from_last)]
                ax.plot(x, y, "bx")
        else:
            ax.plot(ds, fcst[f"yhat{highlight_forecast}"], ls="-", c="b", label=f"yhat{highlight_forecast}")
            ax.plot(ds, fcst[f"yhat{highlight_forecast}"], "bx", label=f"yhat{highlight_forecast}")

            if len(quantiles) > 1:
                for i in range(1, len(quantiles)):
                    ax.fill_between(
                        ds,
                        fcst[f"yhat{highlight_forecast}"],
                        fcst[f"yhat{highlight_forecast} {round(quantiles[i] * 100, 1)}%"],
                        color="#0072B2",
                        alpha=0.2,
                    )

    ax.plot(ds, fcst["y"], "k.", label="actual y")

    # Specify formatting to workaround matplotlib issue #12925
    locator = AutoDateLocator(interval_multiples=False)
    formatter = AutoDateFormatter(locator)
    ax.xaxis.set_major_locator(locator)
    ax.xaxis.set_major_formatter(formatter)
    ax.grid(True, which="major", c="gray", ls="-", lw=1, alpha=0.2)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    handles, labels = ax.axes.get_legend_handles_labels()
    if len(labels) > 10:
        ax.legend(handles[:10] + [handles[-1]], labels[:10] + [labels[-1]])
        log.warning("Legend is available only for the ten first handles")
    else:
        ax.legend(handles, labels)
    fig = fig.tight_layout()
    return fig


def plot_components(
    m,
    fcst,
    plot_configuration,
    df_name="__df__",
    quantile=0.5,
    one_period_per_season=False,
    figsize=None,
):
    """Plot the NeuralProphet forecast components.

    Parameters
    ----------
        m : NeuralProphet
            Fitted model
        fcst : pd.DataFrame
            Output of m.predict
        plot_configuration: dict
            dict of configured components to plot
        df_name : str
            ID from time series that should be plotted
        quantile : float
            Quantile for which the forecast components are to be plotted
        one_period_per_season : bool
            Plot one period per season, instead of the true seasonal components of the forecast.
        figsize : tuple
            Width, height in inches.

            Note
            ----
            Default value is set to ``None`` ->  automatic ``figsize = (10, 3 * npanel)``

    Returns
    -------
        matplotlib.pyplot.figure
            Figure showing the NeuralProphet forecast components
    """
    log.debug("Plotting forecast components")
    fcst = fcst.fillna(value=np.nan)
    components_to_plot = plot_configuration["components_list"]

    # set number of axes based on selected plot_names and sort them according to order in components
    panel_names = list(set(next(iter(dic.values())).lower() for dic in components_to_plot))
    panel_order = [x for dic in components_to_plot for x in panel_names if x in dic["plot_name"].lower()]
    npanel = len(panel_names)
    figsize = figsize if figsize else (10, 3 * npanel)
    fig, axes = plt.subplots(npanel, 1, facecolor="w", figsize=figsize)
    if npanel == 1:
        axes = [axes]
    multiplicative_axes = []
    ax = 0
    # for ax, comp in zip(axes, components):
    for comp in components_to_plot:
        name = comp["plot_name"].lower()
        ax = axes[panel_order.index(name)]
        if (
            name in ["trend"]
            or ("ar" in name and "ahead" in name)
            or ("lagged regressor" in name and "ahead" in name)
            or ("uncertainty" in name)
        ):
            plot_forecast_component(fcst=fcst, ax=ax, **comp)
        elif "event" in name or "future regressor" in name:
            if "multiplicative" in comp.keys() and comp["multiplicative"]:
                multiplicative_axes.append(ax)
            plot_forecast_component(fcst=fcst, ax=ax, **comp)
        elif "season" in name:
            if m.config_seasonality.mode == "multiplicative":
                multiplicative_axes.append(ax)
            if one_period_per_season:
                comp_name = comp["comp_name"]
                if comp_name.lower() == "weekly" or m.config_seasonality.periods[comp_name].period == 7:
                    plot_weekly(m=m, ax=ax, quantile=quantile, comp_name=comp_name, df_name=df_name)
                elif comp_name.lower() == "yearly" or m.config_seasonality.periods[comp_name].period == 365.25:
                    plot_yearly(m=m, ax=ax, quantile=quantile, comp_name=comp_name, df_name=df_name)
                elif comp_name.lower() == "daily" or m.config_seasonality.periods[comp_name].period == 1:
                    plot_daily(m=m, ax=ax, quantile=quantile, comp_name=comp_name, df_name=df_name)
                else:
                    plot_custom_season(m=m, ax=ax, quantile=quantile, comp_name=comp_name, df_name=df_name)
            else:
                comp_name = f"season_{comp['comp_name']}"
                plot_forecast_component(fcst=fcst, ax=ax, comp_name=comp_name, plot_name=comp["plot_name"])
        elif "auto-regression" in name or "lagged regressor" in name:
            plot_multiforecast_component(fcst=fcst, ax=ax, **comp)

    fig = fig.tight_layout()
    # Reset multiplicative axes labels after tight_layout adjustment
    for ax in multiplicative_axes:
        ax = set_y_as_percent(ax)
    return fig


def plot_forecast_component(
    fcst,
    comp_name,
    plot_name=None,
    ax=None,
    figsize=(10, 6),
    multiplicative=False,
    bar=False,
    rolling=None,
    add_x=False,
    fill=False,
):
    """Plot a particular component of the forecast.

    Parameters
    ----------
        fcst : pd.DataFrame
            Output of m.predict
        comp_name : str
            Name of the component to plot
        plot_name : str
            Name of the plot Title
        ax : matplotlib axis
            Matplotlib Axes to plot on
        figsize : tuple
            Width, height in inches. Ignored if ax is not None

            Note
            ----
            Default value is set to ``figsize = (10, 6)``
        multiplicative : bool
            Set y axis as percentage
        bar : bool
            Make barplot
        rolling : int
            Rolling average underplot
        add_x : bool
            Add x symbols to plotted points
        fill: bool
            Add fill between signal and x(y=0) axis

    Returns
    -------
        matplotlib.artist.Artist
            List of Artist objects containing a particular forecast component
    """
    fcst = fcst.fillna(value=np.nan)
    artists = []
    if not ax:
        fig = plt.figure(facecolor="w", figsize=figsize)
        ax = fig.add_subplot(111)
    fcst_t = fcst["ds"].dt.to_pydatetime()
    if rolling is not None:
        rolling_avg = fcst[comp_name].rolling(rolling, min_periods=1, center=True).mean()
        if bar:
            artists += ax.bar(fcst_t, rolling_avg, width=1.00, color="#0072B2", alpha=0.5)
        else:
            artists += ax.plot(fcst_t, rolling_avg, ls="-", color="#0072B2", alpha=0.5)
            if add_x:
                artists += ax.plot(fcst_t, fcst[comp_name], "bx")
    if "uncertainty" in plot_name.lower():
        y = fcst[comp_name].values - fcst["yhat1"].values
        label = comp_name
    else:
        y = fcst[comp_name].values
        label = None
    if bar:
        artists += ax.bar(fcst_t, y, width=1.00, color="#0072B2")
    elif "uncertainty" in plot_name.lower() and fill:
        ax.fill_between(fcst_t, 0, y, alpha=0.2, label=label, color="#0072B2")
    else:
        artists += ax.plot(fcst_t, y, ls="-", c="#0072B2")
        if add_x or sum(fcst[comp_name].notna()) == 1:
            artists += ax.plot(fcst_t, y, "bx")
    # Specify formatting to workaround matplotlib issue #12925
    locator = AutoDateLocator(interval_multiples=False)
    formatter = AutoDateFormatter(locator)
    ax.xaxis.set_major_locator(locator)
    ax.xaxis.set_major_formatter(formatter)
    ax.grid(True, which="major", c="gray", ls="-", lw=1, alpha=0.2)
    ax.set_xlabel("ds")
    if plot_name is None:
        plot_name = comp_name
    ax.set_ylabel(plot_name)
    if multiplicative:
        ax = set_y_as_percent(ax)
    handles, labels = ax.axes.get_legend_handles_labels()
    ax.legend(handles, labels)
    return ax


def plot_multiforecast_component(
    fcst,
    comp_name,
    plot_name=None,
    ax=None,
    figsize=(10, 6),
    multiplicative=False,
    bar=False,
    focus=1,
    num_overplot=None,
):
    """Plot a particular component of the forecast.

    Parameters
    ----------
        fcst : pd.DataFrame
            Output of m.predict.
        comp_name : str
            Name of the component to plot.
        plot_name : str
            Name of the plot Title.
        ax : matplotlib axis
            Matplotlib Axes to plot on.
        figsize : tuple
            Width, height in inches, ignored if ax is not None.

            Note
            ----
            Default value is set to ``figsize = (10, 6)``

        multiplicative : bool
            Set y axis as percentage
        bar : bool
            Make barplot
        focus : int
            Forecast number to portray in detail.
        num_overplot : int
            Overplot all forecasts up to num

            Note
            ----
            Default value is set to ``num_overplot = None`` -> only plot focus

    Returns
    -------
        matplotlib.artist.Artist
            List of Artist objects containing a particular forecast component
    """
    artists = []
    if not ax:
        fig = plt.figure(facecolor="w", figsize=figsize)
        ax = fig.add_subplot(111)
    fcst_t = fcst["ds"].dt.to_pydatetime()
    col_names = [col_name for col_name in fcst.columns if col_name.startswith(comp_name)]
    if num_overplot is not None:
        assert num_overplot <= len(col_names)
        for i in list(range(num_overplot))[::-1]:
            y = fcst[f"{comp_name}{i + 1}"]
            y = y.values
            alpha_min = 0.2
            alpha_softness = 1.2
            alpha = alpha_min + alpha_softness * (1.0 - alpha_min) / (i + 1.0 * alpha_softness)
            y[-1] = 0
            if bar:
                artists += ax.bar(fcst_t, y, width=1.00, color="#0072B2", alpha=alpha)
            else:
                artists += ax.plot(fcst_t, y, ls="-", color="#0072B2", alpha=alpha)
    if num_overplot is None or focus > 1:
        y = fcst[f"{comp_name}{focus}"]
        y = y.values
        y[-1] = 0
        if bar:
            artists += ax.bar(fcst_t, y, width=1.00, color="b")
        else:
            artists += ax.plot(fcst_t, y, ls="-", color="b")
    # Specify formatting to workaround matplotlib issue #12925
    locator = AutoDateLocator(interval_multiples=False)
    formatter = AutoDateFormatter(locator)
    ax.xaxis.set_major_locator(locator)
    ax.xaxis.set_major_formatter(formatter)
    ax.grid(True, which="major", color="gray", ls="-", lw=1, alpha=0.2)
    ax.set_xlabel("ds")
    if plot_name is None:
        plot_name = comp_name
    ax.set_ylabel(plot_name)
    if multiplicative:
        ax = set_y_as_percent(ax)
    return artists


def plot_nonconformity_scores(scores, alpha, q, method):
    """Plot the nonconformity scores as well as the one-sided interval width (q).

    Parameters
    ----------
        scores : dict
            nonconformity scores
        alpha : float or tuple
                user-specified significance level of the prediction interval, float if coverage error spread arbitrarily
                over left and right tails, tuple of two floats for different coverage error over left and right tails
                respectively
        q : float or list
            prediction interval width (or q)
        method : str
            name of conformal prediction technique used

            Options
                * (default) ``naive``: Naive or Absolute Residual
                * ``cqr``: Conformalized Quantile Regression

    Returns
    -------
        matplotlib.pyplot.figure
            Figure showing the nonconformity score with horizontal line for q-value based on the significance level or
            alpha
    """
    if not isinstance(q, list):
        q_sym = q
        scores = scores["noncon_scores"]
        confidence_levels = np.arange(len(scores)) / len(scores)
        fig, ax = plt.subplots()
        ax.plot(confidence_levels, scores, label="score")
        ax.axvline(x=1 - alpha, color="g", linestyle="-", label=f"(1-alpha) = {1 - alpha}", linewidth=1)
        ax.axhline(y=q, color="r", linestyle="-", label=f"q1 = {round(q_sym, 2)}", linewidth=1)
    else:
        q_lo, q_hi = q
        scores_lo = scores["noncon_scores_lo"]
        scores_hi = scores["noncon_scores_hi"]
        alpha_lo, alpha_hi = alpha
        confidence_levels = np.arange(len(scores_lo)) / len(scores_lo)
        fig, ax = plt.subplots()
        ax.plot(confidence_levels, scores_lo, label="lower score")
        ax.plot(confidence_levels, scores_hi, label="upper score")
        ax.axvline(
            x=1 - alpha_lo,
            color="darkgreen",
            linestyle="-",
            label=f"(1-alpha_lo) = {round(1.0-alpha_lo, 10)}",
            linewidth=1,
        )
        ax.axvline(
            x=1 - alpha_hi,
            color="lightgreen",
            linestyle="-",
            label=f"(1-alpha_hi) = {round(1.0-alpha_hi, 10)}",
            linewidth=1,
        )
        ax.axhline(y=q_lo, color="darkred", linestyle="-", label=f"q1 = {round(q_lo, 2)}", linewidth=1)
        ax.axhline(y=q_hi, color="red", linestyle="-", label=f"q2 = {round(q_hi, 2)}", linewidth=1)
    ax.set_title(f"{method} One-Sided Interval Width with q")
    ax.set_xlabel("Confidence Level")
    ax.set_ylabel("One-Sided Interval Width")
    ax.legend()
    return fig


def plot_interval_width_per_timestep(q_hats, method):
    """Plot the nonconformity scores as well as the one-sided interval width (q).

    Parameters
    ----------
        q_hats : dataframe
            prediction interval widths (or q) for each timestep, contains column ``q_hat_sym`` for symmetric q or
            ``q_hat_lo`` and ``q_hat_hi`` for asymmetric q
        method : str
            name of conformal prediction technique used

            Options
                * (default) ``naive``: Naive or Absolute Residual
                * ``cqr``: Conformalized Quantile Regression

    Returns
    -------
        matplotlib.pyplot.figure
            Figure showing the q-values for each timestep
    """
    fig, ax = plt.subplots()
    # check if q_hats contains q_hat_sym
    if "q_hat_sym" in q_hats.columns:
        q_hats_sym = q_hats["q_hat_sym"]
        ax.plot(range(1, len(q_hats_sym) + 1), q_hats_sym)
    else:
        q_hats_lo = q_hats["q_hat_lo"]
        q_hats_hi = q_hats["q_hat_hi"]
        ax.plot(range(1, len(q_hats_lo) + 1), q_hats_lo, label="lower q")
        ax.plot(range(1, len(q_hats_hi) + 1), q_hats_hi, label="upper q")
    ax.set_title(f"{method} One-Sided Interval Width with q per Timestep")
    ax.set_xlabel("Timestep Number")
    ax.set_ylabel("One-Sided Interval Width")
    # ax.set_xlim(left=1)
    ax.set_ylim(bottom=0)
    return fig



import logging

from neuralprophet import utils
from neuralprophet.components import BaseComponent

log = logging.getLogger("NP.future_regressors")


class FutureRegressors(BaseComponent):
    def __init__(self, config, id_list, quantiles, n_forecasts, device, config_trend_none_bool):
        super().__init__(n_forecasts=n_forecasts, quantiles=quantiles, id_list=id_list, device=device)

        self.config_regressors = config  # config_regressors
        self.regressors_dims = utils.config_regressors_to_model_dims(config)  # config_regressors
        if self.regressors_dims is not None:
            self.n_additive_regressor_params = 0
            self.n_multiplicative_regressor_params = 0
            for name, configs in self.regressors_dims.items():
                if configs["mode"] not in ["additive", "multiplicative"]:
                    log.error("Regressors mode {} not implemented. Defaulting to 'additive'.".format(configs["mode"]))
                    self.regressors_dims[name]["mode"] = "additive"
                if configs["mode"] == "additive":
                    self.n_additive_regressor_params += 1
                elif configs["mode"] == "multiplicative":
                    if config_trend_none_bool:
                        log.error("Multiplicative regressors require trend.")
                        raise ValueError
                    self.n_multiplicative_regressor_params += 1

        else:
            self.config_regressors = None



import datetime
import logging

import numpy as np
import pandas as pd

from neuralprophet.plot_utils import predict_one_season, predict_season_from_dates

log = logging.getLogger("NP.plotly")

try:
    import plotly.graph_objs as go
    from plotly.subplots import make_subplots
    from plotly_resampler import register_plotly_resampler, unregister_plotly_resampler
except ImportError:
    log.error("Importing plotly failed. Interactive plots will not work.")

# UI Configuration
color = "#2d92ff"
xaxis_args = {
    "showline": True,
    "mirror": True,
    "linewidth": 1.5,
}
yaxis_args = {
    "showline": True,
    "mirror": True,
    "linewidth": 1.5,
}
layout_args = {
    "autosize": True,
    "template": "plotly_white",
    "margin": go.layout.Margin(l=0, r=10, b=0, t=10, pad=0),
    "font": dict(size=10),
    "title": dict(font=dict(size=12)),
    "hovermode": "x unified",
}
register_plotly_resampler(mode="auto")


def get_dynamic_axis_range(df_range, type, pad=0.05, inverse=False):
    """Adds a percentage of values at both ends of a list for plotting.

    Parameters
    ----------
        df_range: np.array
            Array of axis values to pad
        type : str
            Type of values in the list to pad
        pad : float
            Percentage of padding to add to each end of the range
        inverse : bool
            Flag for list sorted in an inverted order

    Returns
    -------
        Padded range of values
    """
    if inverse:
        df_range = df_range[::-1]
    delta = df_range[round(len(df_range) * pad)]
    if type == "dt":
        range_min = min(df_range) + (min(df_range) - delta)
        range_max = max(df_range) + (delta - min(df_range))
    elif type == "numeric":
        range_min = min(df_range) - delta
        range_max = max(df_range) + delta
    else:
        raise NotImplementedError(f"The type {type} is not implemented.")
    if inverse:
        df_range = df_range[::-1]
    return [range_min, range_max]


def plot_trend_change(m, quantile, plot_name="Trend Change", df_name="__df__"):
    """Make a barplot of the magnitudes of trend-changes.

    Parameters
    ----------
        m : NeuralProphet
            Fitted model
        quantile : float
            The quantile for which the yearly seasonality is plotted
        plot_name : str
            Name of the plot Title
        df_name : str
            Name of dataframe to refer to data params from original keys of train dataframes

            Note
            ----
            Only used for local normalization in global modeling

    Returns
    -------
        Dictionary with plotly traces, xaxis and yaxis
    """
    if isinstance(df_name, list):
        df_name = df_name[0]
    data_params = m.config_normalization.get_data_params(df_name)
    start = data_params["ds"].shift
    scale = data_params["ds"].scale
    time_span_seconds = scale.total_seconds()
    cp_t = np.array([])
    for cp in m.model.config_trend.changepoints:
        cp_t = np.append(cp_t, start + datetime.timedelta(seconds=cp * time_span_seconds))
    # Global/Local Mode
    if m.model.config_trend.trend_global_local == "local":
        quantile_index = m.model.quantiles.index(quantile)
        weights = m.model.trend.get_trend_deltas.detach()[quantile_index, m.model.id_dict[df_name], :].numpy()
    else:
        quantile_index = m.model.quantiles.index(quantile)
        weights = m.model.trend.get_trend_deltas.detach()[quantile_index, 0, :].numpy()
    # add end-point to force scale to match trend plot
    cp_t = np.append(cp_t, start + scale)
    weights = np.append(weights, [0.0])

    traces = []
    traces.append(
        go.Bar(
            name=plot_name,
            x=cp_t,
            y=weights,
            marker_color=color,
        ),
    )

    padded_range = get_dynamic_axis_range(cp_t, type="dt")
    xaxis = go.layout.XAxis(
        title="Trend segment",
        type="date",
        range=padded_range,
    )
    yaxis = go.layout.YAxis(
        rangemode="normal",
        title=go.layout.yaxis.Title(text=plot_name),
    )

    return {"traces": traces, "xaxis": xaxis, "yaxis": yaxis}


def plot_trend(m, quantile, plot_name="Trend Change", df_name="__df__"):
    """Make a barplot of the magnitudes of trend-changes.

    Parameters
    ----------
        m : NeuralProphet
            Fitted model
        quantile : float
            The quantile for which the yearly seasonality is plotted
        plot_name : str
            Name of the plot Title
        df_name : str
            Name of dataframe to refer to data params from original keys of train dataframes

            Note
            ----
            Only used for local normalization in global modeling

    Returns
    -------
        Dictionary with plotly traces, xaxis and yaxis
    """
    traces = []
    line_width = 2

    if m.config_trend.n_changepoints == 0:
        if isinstance(df_name, list):
            df_name = df_name[0]
        data_params = m.config_normalization.get_data_params(df_name)
        t_start = data_params["ds"].shift
        t_end = t_start + data_params["ds"].scale
        quantile_index = m.model.quantiles.index(quantile)

        fcst_t = pd.Series([t_start, t_end]).dt.to_pydatetime()
        trend_0 = m.model.trend.bias[quantile_index].detach().numpy().squeeze().reshape(1)
        if m.config_trend.growth == "off":
            trend_1 = trend_0
        else:
            if m.model.config_trend.trend_global_local == "local":
                trend_1 = trend_0 + m.model.trend.trend_k0[quantile_index, m.model.id_dict[df_name]].detach().numpy()
            else:
                trend_1 = trend_0 + m.model.trend.trend_k0[quantile_index, 0].detach().numpy()

        data_params = m.config_normalization.get_data_params(df_name)
        shift = data_params["y"].shift
        scale = data_params["y"].scale
        trend_0 = trend_0 * scale + shift
        trend_1 = trend_1 * scale + shift

        traces.append(
            go.Scatter(
                name=plot_name,
                x=fcst_t,
                y=np.concatenate([trend_0, trend_1]),
                mode="lines",
                line=dict(color=color, width=line_width),
                fill="none",
            )
        )
        extended_daterange = pd.date_range(
            start=fcst_t[0].strftime("%Y-%m-%d"), end=fcst_t[1].strftime("%Y-%m-%d")
        ).to_pydatetime()
        padded_range = get_dynamic_axis_range(extended_daterange, type="dt")
    else:
        mean_std = True
        if not isinstance(df_name, list):
            df_name = [df_name]
            # if global df with no specified df_name: plot mean and std, otherwise: don't
            mean_std = False
        df_y = pd.DataFrame()
        for df_name_i in df_name:
            data_params = m.config_normalization.get_data_params(df_name_i)
            t_start = data_params["ds"].shift
            t_end = t_start + data_params["ds"].scale
            quantile_index = m.model.quantiles.index(quantile)

            days = pd.date_range(start=t_start, end=t_end, freq=m.data_freq)
            df_i = pd.DataFrame({"ds": days})
            df_i["ID"] = df_name_i
            df_y = pd.concat((df_y, df_i), ignore_index=True)

        df_trend = m.predict_trend(df=df_y, quantile=quantile)
        if mean_std:
            df_trend_q90 = df_trend.groupby("ds")[["trend"]].apply(lambda x: x.quantile(0.9))
            df_trend_q10 = df_trend.groupby("ds")[["trend"]].apply(lambda x: x.quantile(0.1))
            df_trend = df_trend.groupby("ds")[["trend"]].apply(lambda x: x.mean())
            df_trend["ID"] = m.id_list[0]
            df_y = df_y[df_y["ID"] == m.id_list[0]]

        traces.append(
            go.Scatter(
                name=plot_name + " mean" if mean_std else plot_name,
                x=df_y["ds"].dt.to_pydatetime(),
                y=df_trend["trend"],
                mode="lines",
                line=dict(color=color, width=line_width),
                fill="none",
            )
        )
        if mean_std:
            # If more than on ID has been provided, and no df_name has been specified: plot mean and quants of the
            # component
            filling = "tonexty"
            traces.append(
                go.Scatter(
                    name="Quants: 10%",
                    x=df_y["ds"].dt.to_pydatetime(),
                    y=df_trend_q10["trend"],
                    mode="lines",
                    line=dict(color="rgba(45, 146, 255, 0.2)", width=1),
                    fillcolor="rgba(45, 146, 255, 0.2)",
                    showlegend=True,
                )
            )
            traces.append(
                go.Scatter(
                    name="Quants: 90%",
                    x=df_y["ds"].dt.to_pydatetime(),
                    y=df_trend_q90["trend"],
                    fill=filling,
                    mode="lines",
                    line=dict(color="rgba(45, 146, 255, 0.2)", width=1),
                    fillcolor="rgba(45, 146, 255, 0.2)",
                    showlegend=True,
                )
            )
        padded_range = get_dynamic_axis_range(df_y["ds"].dt.to_pydatetime(), type="dt")

    xaxis = go.layout.XAxis(
        title="ds",
        type="date",
        range=padded_range,
    )
    yaxis = go.layout.YAxis(
        rangemode="normal",
        title=go.layout.yaxis.Title(text=plot_name),
    )

    return {"traces": traces, "xaxis": xaxis, "yaxis": yaxis}


def plot_scalar_weights(weights, plot_name, focus=None, multiplicative=False):
    """Make a barplot of the regressor weights.

    Parameters
    ----------
        weights : list
            tuples of (name, weights)
        plot_name : str
            Name of the plot Title
        focus : int
            Show weights for this forecast, if provided
        multiplicative : bool
            Flag to set y axis as percentage

    Returns
    -------
        Dictionary with Plotly traces, xaxis and yaxis
    """
    traces = []

    names = []
    values = []
    for name, weights in weights:
        names.append(name)
        weight = np.squeeze(weights)
        if len(weight.shape) > 1:
            raise ValueError("Not scalar " + plot_name)
        if len(weight.shape) == 1 and len(weight) > 1:
            if focus is not None:
                weight = weight[focus - 1]
            else:
                weight = np.mean(weight)
        values.append(weight)

    traces.append(
        go.Bar(
            name=plot_name,
            x=np.array(names),
            y=values,
            marker_color=color,
            width=0.8,
        )
    )

    xaxis = go.layout.XAxis(title=f"{plot_name} name")

    if "lagged" in plot_name.lower():
        if focus is None:
            yaxis = go.layout.YAxis(
                rangemode="normal",
                title=go.layout.yaxis.Title(text=f"{plot_name} weight (avg)"),
            )
        else:
            yaxis = go.layout.YAxis(
                rangemode="normal",
                title=go.layout.yaxis.Title(text=f"{plot_name} weight ({focus})-ahead"),
            )
    else:
        yaxis = go.layout.YAxis(
            rangemode="normal",
            title=go.layout.yaxis.Title(text=f"{plot_name} weight"),
        )

    if multiplicative:
        yaxis.update(tickformat=".1%", hoverformat=".4%")

    return {"traces": traces, "xaxis": xaxis, "yaxis": yaxis}


def plot_lagged_weights(weights, comp_name, focus=None):
    """Make a barplot of the importance of lagged inputs.

    Parameters
    ----------
        weights : list
            tuples of (name, weights)
        comp_name : str
            Name of lagged inputs
        focus : int
            Show weights for this forecast, if provided

    Returns
    -------
        Dictionary with plotly traces, xaxis and yaxis
    """
    traces = []

    n_lags = weights.shape[1]
    lags_range = np.array(range(1, 1 + n_lags))[::-1]
    if focus is None:
        weights = np.sum(np.abs(weights), axis=0)
        weights = weights / np.sum(weights)

        traces.append(
            go.Bar(
                name=comp_name,
                x=lags_range,
                y=weights,
                marker_color=color,
            )
        )

    else:
        if len(weights.shape) == 2:
            weights = weights[focus - 1, :]

        traces.append(go.Bar(name=comp_name, x=lags_range, y=weights, marker_color=color, width=0.8))

    padded_range = get_dynamic_axis_range(lags_range, type="numeric", inverse=True)
    xaxis = go.layout.XAxis(title=f"{comp_name} lag number", range=padded_range)

    if focus is None:
        yaxis = go.layout.YAxis(
            rangemode="normal",
            title=go.layout.yaxis.Title(text=f"{comp_name} relevance"),
            tickformat=",.0%",
        )
        # ax = set_y_as_percent(ax)
    else:
        yaxis = go.layout.YAxis(
            rangemode="normal",
            title=go.layout.yaxis.Title(text=f"{comp_name} weight ({focus})-ahead"),
        )

    return {"traces": traces, "xaxis": xaxis, "yaxis": yaxis}


def plot_yearly(m, quantile, comp_name="yearly", yearly_start=0, quick=True, multiplicative=False, df_name="__df__"):
    """Plot the yearly component of the forecast.

    Parameters
    ----------
        m : NeuralProphet
            Fitted model
        quantile : float
            The quantile for which the yearly seasonality is plotted
        comp_name : str
            Name of seasonality component
        yearly_start : int
            Specifying the start day of the yearly seasonality plot

            Options
                * (default) ``yearly_start = 0``: starts the year on Jan 1
                * ``yearly_start = 1``: shifts by 1 day to Jan 2, and so on
        quick : bool
            Use quick low-level call of model
        multiplicative : bool
            Flag to set y axis as percentage
        df_name : str
            Name of dataframe to refer to data params from original keys of train dataframes

            Note
            ----
            Only used for local normalization in global modeling

    """
    traces = []
    line_width = 2

    # Compute yearly seasonality for a Jan 1 - Dec 31 sequence of dates.
    days = pd.date_range(start="2017-01-01", periods=365) + pd.Timedelta(days=yearly_start)
    df_y = pd.DataFrame({"ds": days})
    if not isinstance(df_name, list):
        df_y["ID"] = df_name
    mean_std = False  # Indicates whether mean and std of global df shall be plotted
    if isinstance(df_name, list):
        mean_std = True
        quick = False
        df_y = pd.DataFrame()
        for i in range(m.id_list.__len__()):
            df_i = pd.DataFrame({"ds": days})
            df_i["ID"] = m.id_list[i]
            df_y = pd.concat((df_y, df_i), ignore_index=True)
    if quick:
        predicted = predict_season_from_dates(m, dates=df_y["ds"], name=comp_name, quantile=quantile, df_name=df_name)
    else:
        predicted = m.predict_seasonal_components(df_y, quantile=quantile)[["ds", "ID", comp_name]]

    if mean_std:
        # If more than on ID has been provided, and no df_name has been specified: plot median and quants across all IDs
        predicted_q90 = predicted[["ds", comp_name]].groupby("ds").apply(lambda x: x.quantile(0.9))
        predicted_q10 = predicted[["ds", comp_name]].groupby("ds").apply(lambda x: x.quantile(0.1))
        predicted = predicted[["ds", comp_name]].groupby("ds").apply(lambda x: x.mean())
        predicted["ID"] = m.id_list[0]
        df_y = df_y[df_y["ID"] == m.id_list[0]]

    traces.append(
        go.Scatter(
            name=comp_name + " Mean" if mean_std else comp_name,
            x=df_y["ds"].dt.to_pydatetime(),
            y=predicted[comp_name],
            mode="lines",
            line=dict(color=color, width=line_width),
            fill="none",
        )
    )
    if mean_std:
        filling = "tonexty"
        traces.append(
            go.Scatter(
                name="Quant 10%",
                x=df_y["ds"],
                y=predicted_q10[comp_name],
                mode="lines",
                line=dict(color="rgba(45, 146, 255, 0.2)", width=1),
                fillcolor="rgba(45, 146, 255, 0.2)",
                showlegend=True,
            )
        )
        traces.append(
            go.Scatter(
                name="Quant 90%",
                x=df_y["ds"],
                y=predicted_q90[comp_name],
                fill=filling,
                mode="lines",
                line=dict(color="rgba(45, 146, 255, 0.2)", width=1),
                fillcolor="rgba(45, 146, 255, 0.2)",
                showlegend=False,
            )
        )

    padded_range = get_dynamic_axis_range(df_y["ds"].dt.to_pydatetime(), type="dt")
    xaxis = go.layout.XAxis(
        title="Day of year",
        range=padded_range,
        tickformat="%B %e",
    )
    yaxis = go.layout.YAxis(
        rangemode="normal",
        title=go.layout.yaxis.Title(text=f"Seasonality: {comp_name}"),
    )

    if multiplicative:
        yaxis.update(tickformat=".1%", hoverformat=".4%")

    return {"traces": traces, "xaxis": xaxis, "yaxis": yaxis}


def plot_weekly(m, quantile, comp_name="weekly", weekly_start=0, quick=True, multiplicative=False, df_name="__df__"):
    """Plot the weekly component of the forecast.

    Parameters
    ----------
        m : NeuralProphet
            Fitted model
        quantile : float
            The quantile for which the yearly seasonality is plotted
        comp_name : str
            Name of seasonality component
        weekly_start : int
            Specifying the start day of the weekly seasonality plot

            Options
                * (default) ``weekly_start = 0``: starts the week on Sunday
                * ``weekly_start = 1``: shifts by 1 day to Monday, and so on
        quick : bool
            Use quick low-level call of model
        multplicative : bool
            Flag to set y axis as percentage
        df_name : str
            Name of dataframe to refer to data params from original keys of train dataframes

            Note
            ----
            Only used for local normalization in global modeling
    Returns
    -------
        Dictionary with plotly traces, xaxis and yaxis

    """
    traces = []
    line_width = 2

    week_days = 7
    if m.data_freq == "B":
        week_days = 5
        weekly_start = 1
    days_i = pd.date_range(start="2017-01-01", periods=week_days * 24, freq="H") + pd.Timedelta(days=weekly_start)
    df_w = pd.DataFrame({"ds": days_i})
    if not isinstance(df_name, list):
        df_w["ID"] = df_name
    mean_std = False  # Indicates whether mean and std of global df shall be plotted
    if isinstance(df_name, list):
        df_w = pd.DataFrame()
        quick = False
        mean_std = True
        for i in range(m.id_list.__len__()):
            df_i = pd.DataFrame({"ds": days_i})
            df_i["ID"] = m.id_list[i]
            df_w = pd.concat((df_w, df_i), ignore_index=True)
    if quick:
        predicted = predict_season_from_dates(m, dates=df_w["ds"], name=comp_name, quantile=quantile, df_name=df_name)
    else:
        predicted = m.predict_seasonal_components(df_w, quantile=quantile)[["ds", "ID", comp_name]]
    days = pd.date_range(start="2017-01-01", periods=week_days) + pd.Timedelta(days=weekly_start)

    if mean_std:
        # If more than on ID has been provided, and no df_name has been specified: plot median and quants across all IDs
        predicted_q90 = predicted.groupby("ds")[[comp_name]].apply(lambda x: x.quantile(0.9))
        predicted_q10 = predicted.groupby("ds")[[comp_name]].apply(lambda x: x.quantile(0.1))
        predicted = predicted.groupby("ds")[[comp_name]].apply(lambda x: x.mean())
        predicted["ID"] = m.id_list[0]
        df_w = df_w[df_w["ID"] == m.id_list[0]]

    days = pd.date_range(start="2017-01-01", periods=8) + pd.Timedelta(days=weekly_start)
    days = days.day_name()

    traces.append(
        go.Scatter(
            name=comp_name + " Mean" if mean_std else comp_name,
            x=np.array(range(len(days_i))),
            # x=df_w['ds'].dt.to_pydatetime(),
            y=predicted[comp_name],
            mode="lines",
            line=dict(color=color, width=line_width),
            fill="none",
        )
    )
    if mean_std:
        filling = "tonexty"
        traces.append(
            go.Scatter(
                name="Quant 10%",
                x=np.array(range(len(days_i))),
                y=predicted_q10[comp_name],
                mode="lines",
                line=dict(color="rgba(45, 146, 255, 0.2)", width=1),
                fillcolor="rgba(45, 146, 255, 0.2)",
                showlegend=True,
            )
        )
        traces.append(
            go.Scatter(
                name="Quant 90%",
                x=np.array(range(len(days_i))),
                y=predicted_q90[comp_name],
                fill=filling,
                mode="lines",
                line=dict(color="rgba(45, 146, 255, 0.2)", width=1),
                fillcolor="rgba(45, 146, 255, 0.2)",
                showlegend=False,
            )
        )
    padded_range = get_dynamic_axis_range(list(range(len(days_i))), type="numeric")
    xaxis = go.layout.XAxis(
        title="Day of week",
        tickmode="array",
        range=padded_range,
        tickvals=[x * 24 for x in range(len(days) + 1 - weekly_start)],
        ticktext=list(days) + [days[0]] if m.data_freq != "B" else list(days),
    )
    yaxis = go.layout.YAxis(
        rangemode="normal",
        title=go.layout.yaxis.Title(text=f"Seasonality: {comp_name}"),
    )

    if multiplicative:
        yaxis.update(tickformat=".1%", hoverformat=".4%")

    return {"traces": traces, "xaxis": xaxis, "yaxis": yaxis}


def plot_daily(m, quantile, comp_name="daily", quick=True, multiplicative=False, df_name="__df__"):
    """Plot the daily component of the forecast.

    Parameters
    ----------
        m : NeuralProphet
            Fitted model
        quantile : float
            The quantile for which the yearly seasonality is plotted
        comp_name : str
            Name of seasonality component if previously changed from default ``daily``
        quick : bool
            Use quick low-level call of model
        ax : matplotlib axis
            Matplotlib Axes to plot on
        multiplicative: bool
            Flag whether to set y axis as percentage
        df_name : str
            Name of dataframe to refer to data params from original keys of train dataframes

            Note
            ----
            Only used for local normalization in global modeling

    Returns
    -------
        Dictionary with plotly traces, xaxis and yaxis
    """
    traces = []
    line_width = 2

    # Compute daily seasonality
    days_i = pd.date_range(start="2017-01-01", periods=24 * 12, freq="5min")
    df_d = pd.DataFrame({"ds": days_i})
    if not isinstance(df_name, list):
        df_d["ID"] = df_name
    mean_std = False  # Indicates whether mean and std of global df shall be plotted
    if isinstance(df_name, list):
        df_d = pd.DataFrame()
        quick = False
        mean_std = True
        for i in range(m.id_list.__len__()):
            df_i = pd.DataFrame({"ds": days_i})
            df_i["ID"] = m.id_list[i]
            df_d = pd.concat((df_d, df_i), ignore_index=True)
    if quick:
        predicted = predict_season_from_dates(m, dates=df_d["ds"], name=comp_name, quantile=quantile, df_name=df_name)
    else:
        predicted = m.predict_seasonal_components(df_d, quantile=quantile)[["ds", "ID", comp_name]]

    if mean_std:
        # If more than on ID has been provided, and no df_name has been specified: plot median and quants across all IDs
        predicted_q90 = predicted.groupby("ds")[[comp_name]].apply(lambda x: x.quantile(0.9))
        predicted_q10 = predicted.groupby("ds")[[comp_name]].apply(lambda x: x.quantile(0.1))
        predicted = predicted.groupby("ds")[[comp_name]].apply(lambda x: x.mean())
        predicted["ID"] = m.id_list[0]
        df_d = df_d[df_d["ID"] == m.id_list[0]]

    traces.append(
        go.Scatter(
            name=comp_name + " Mean" if mean_std else comp_name,
            x=np.array(range(len(days_i))),
            # x=df_d['ds'].dt.to_pydatetime(),
            y=predicted[comp_name],
            mode="lines",
            line=dict(color=color, width=line_width),
            fill="none",
        ),
    )
    if mean_std:
        filling = "tonexty"
        traces.append(
            go.Scatter(
                name="Quant 10%",
                x=np.array(range(len(days_i))),
                y=predicted_q10[comp_name],
                mode="lines",
                line=dict(color="rgba(45, 146, 255, 0.2)", width=1),
                fillcolor="rgba(45, 146, 255, 0.2)",
                showlegend=True,
            )
        )
        traces.append(
            go.Scatter(
                name="Quant 90%",
                x=np.array(range(len(days_i))),
                y=predicted_q90[comp_name],
                fill=filling,
                mode="lines",
                line=dict(color="rgba(45, 146, 255, 0.2)", width=1),
                fillcolor="rgba(45, 146, 255, 0.2)",
                showlegend=False,
            )
        )
    padded_range = get_dynamic_axis_range(list(range(len(days_i))), type="numeric")
    xaxis = go.layout.XAxis(
        title="Hour of day",
        tickmode="array",
        range=padded_range,
        tickvals=list(np.arange(25) * 12),
        ticktext=list(np.arange(25)),
    )
    yaxis = go.layout.YAxis(
        rangemode="normal",
        title=go.layout.yaxis.Title(text=f"Seasonality: {comp_name}"),
    )

    if multiplicative:
        yaxis.update(tickformat=".1%", hoverformat=".4%")

    return {"traces": traces, "xaxis": xaxis, "yaxis": yaxis}


def plot_custom_season(m, comp_name, quantile, multiplicative=False, df_name="__df__"):
    """Plot any seasonal component of the forecast.

    Parameters
    ----------
        m : NeuralProphet
            Fitted model
        comp_name : str
            Name of seasonality component
        quantile : float
            The quantile for which the custom season is plotted
        multiplicative : bool
            Flag whether to set y axis as percentage
        df_name : str
            Name of dataframe to refer to data params from original keys of train dataframes

            Note
            ----
            Only used for local normalization in global modeling
    Returns
    -------
        Dictionary with plotly traces, xaxis and yaxis

    """

    traces = []
    line_width = 2

    t_i, predicted = predict_one_season(m, quantile=quantile, name=comp_name, n_steps=300, df_name=df_name)
    traces = []

    traces.append(
        go.Scatter(
            name=comp_name,
            x=t_i,
            y=predicted,
            mode="lines",
            line=dict(color=color, width=line_width),
            fill="none",
        )
    )
    padded_range = get_dynamic_axis_range(t_i, type="numeric")
    xaxis = go.layout.XAxis(
        title=f"One period: {comp_name}",
        range=padded_range,
    )
    yaxis = go.layout.YAxis(
        rangemode="normal",
        title=go.layout.yaxis.Title(text=f"Seasonality: {comp_name}"),
    )

    if multiplicative:
        yaxis.update(tickformat=".1%", hoverformat=".4%")

    return {"traces": traces, "xaxis": xaxis, "yaxis": yaxis}


def plot_parameters(
    m,
    plot_configuration,
    quantile=0.5,
    weekly_start=0,
    yearly_start=0,
    figsize=(700, 210),
    df_name=None,
    forecast_in_focus=None,
    resampler_active=False,
    plotly_static=False,
):
    """Plot the parameters that the model is composed of, visually.

    Parameters
    ----------
        m : NeuralProphet
            Fitted model
        plot_configuration: dict
            dict of configured parameters to plot
        quantile : float
            The quantile for which the model parameters are to be plotted
        weekly_start : int
            Specifying the start day of the weekly seasonality plot

            Options
                * (default) ``weekly_start = 0``: starts the week on Sunday
                * ``weekly_start = 1``: shifts by 1 day to Monday, and so on
        yearly_start : int
            Specifying the start day of the yearly seasonality plot.

            Options
                * (default) ``yearly_start = 0``: starts the year on Jan 1
                * ``yearly_start = 1``: shifts by 1 day to Jan 2, and so on
        figsize : tuple
            Width, height in inches.

            Note
            ----
            Default value is set to ``None`` ->  automatic ``figsize = (700, 210 * npanel)``
        df_name : str
            Name of dataframe to refer to data params from original keys of train dataframes

            Note
            ----
            Only used for local normalization in global modeling
        forecast_in_focus: int
            optinal, i-th step ahead forecast to plot

            Note
            ----
            None (default): plot self.highlight_forecast_step_n by default
        resampler_active : bool
            Flag whether to activate the plotly-resampler
        plotly_static: bool
            Flag whether to generate a static svg image

    Returns:
        Plotly figure
    """
    if resampler_active:
        register_plotly_resampler(mode="auto")
    else:
        unregister_plotly_resampler()
    compnents_to_plot = plot_configuration["components_list"]
    additive_future_regressors = plot_configuration["additive_future_regressors"]
    additive_events = plot_configuration["additive_events"]
    multiplicative_future_regressors = plot_configuration["multiplicative_future_regressors"]
    multiplicative_events = plot_configuration["multiplicative_events"]
    lagged_scalar_regressors = plot_configuration["lagged_scalar_regressors"]

    npanel = len(compnents_to_plot)
    figsize = figsize if figsize else (700, 210 * npanel)

    # Create Plotly subplot figure and add the components to it
    fig = make_subplots(npanel, cols=1, print_grid=False)
    fig.update_layout(go.Layout(showlegend=False, width=figsize[0], height=figsize[1] * npanel, **layout_args))

    for i, comp in enumerate(compnents_to_plot):
        is_multiplicative = False
        plot_name = comp["plot_name"].lower()
        if plot_name.startswith("trend"):
            if "change" in plot_name:
                trace_object = plot_trend_change(m, quantile=quantile, plot_name=comp["plot_name"], df_name=df_name)
            else:
                trace_object = plot_trend(m, quantile=quantile, plot_name=comp["plot_name"], df_name=df_name)

        elif plot_name.startswith("seasonality"):
            name = comp["comp_name"]
            if m.config_seasonality.mode == "multiplicative":
                is_multiplicative = True
            if name.lower() == "weekly" or m.config_seasonality.periods[name].period == 7:
                trace_object = plot_weekly(
                    m=m,
                    quantile=quantile,
                    weekly_start=weekly_start,
                    comp_name=name,
                    multiplicative=is_multiplicative,
                    df_name=df_name,
                )
            elif name.lower() == "yearly" or m.config_seasonality.periods[name].period == 365.25:
                trace_object = plot_yearly(
                    m=m,
                    quantile=quantile,
                    yearly_start=yearly_start,
                    comp_name=name,
                    multiplicative=is_multiplicative,
                    df_name=df_name,
                )
            elif name.lower() == "daily" or m.config_seasonality.periods[name].period == 1:
                trace_object = plot_daily(
                    m=m, quantile=quantile, comp_name=name, multiplicative=is_multiplicative, df_name=df_name
                )
            else:
                trace_object = plot_custom_season(
                    m=m, quantile=quantile, comp_name=name, multiplicative=is_multiplicative, df_name=df_name
                )
        elif plot_name == "lagged weights":
            trace_object = plot_lagged_weights(
                weights=comp["weights"], comp_name=comp["comp_name"], focus=comp["focus"]
            )
        else:
            if plot_name == "additive future regressor":
                weights = additive_future_regressors
            elif plot_name == "multiplicative future regressor":
                is_multiplicative = True
                weights = multiplicative_future_regressors
            elif plot_name == "lagged scalar regressor":
                weights = lagged_scalar_regressors
            elif plot_name == "additive event":
                weights = additive_events
            elif plot_name == "multiplicative event":
                is_multiplicative = True
                weights = multiplicative_events
            trace_object = plot_scalar_weights(
                weights=weights, plot_name=comp["plot_name"], focus=forecast_in_focus, multiplicative=is_multiplicative
            )

        if i == 0:
            xaxis = fig["layout"]["xaxis"]
            yaxis = fig["layout"]["yaxis"]
        else:
            xaxis = fig["layout"][f"xaxis{i + 1}"]
            yaxis = fig["layout"][f"yaxis{i + 1}"]

        xaxis.update(trace_object["xaxis"])
        yaxis.update(trace_object["yaxis"])
        xaxis.update(**xaxis_args)
        yaxis.update(**yaxis_args)
        for trace in trace_object["traces"]:
            fig.add_trace(trace, row=i + 1, col=1)  # adapt var name to plotly-resampler
        unregister_plotly_resampler()
    return fig



import logging
from typing import List, Optional

import numpy as np
import pandas as pd

from neuralprophet import df_utils, time_dataset
from neuralprophet.configure import (
    ConfigCountryHolidays,
    ConfigEvents,
    ConfigFutureRegressors,
    ConfigLaggedRegressors,
    ConfigSeasonality,
)
from neuralprophet.np_types import Components

log = logging.getLogger("NP.data.processing")


def _reshape_raw_predictions_to_forecst_df(
    df: pd.DataFrame,
    predicted: np.ndarray,
    components: Optional[Components],
    prediction_frequency: Optional[dict],
    dates: pd.Series,
    n_forecasts: int,
    max_lags: int,
    freq: Optional[str],
    quantiles: List[float],
    config_lagged_regressors: Optional[ConfigLaggedRegressors],
) -> pd.DataFrame:
    """
    Turns forecast-origin-wise predictions into forecast-target-wise predictions.

    Parameters
    ----------
        df : pd.DataFrame
            input dataframe
        predicted : np.array
            Array containing the forecasts
        components : dict[np.array]
            Dictionary of components containing an array of each components' contribution to the forecast
        prediction_frequency : str
            Frequency of the predictions
        dates : pd.Series
            timestamps referring to the start of the predictions
        n_forecasts : int
            Number of steps ahead of prediction time step to forecast.
        max_lags : int
            Maximum number of lags to use
        freq : str
            Data step sizes. Frequency of data recording.
        quantiles : list[float]
            List of quantiles to include in the forecast
        config_lagged_regressors : ConfigLaggedRegressors
            Configuration for lagged regressors

    Returns
    -------
        pd.DataFrame
            columns ``ds``, ``y``, ``trend`` and [``yhat<i>``]

            Note
            ----
            where yhat<i> refers to the i-step-ahead prediction for this row's datetime.
            e.g. yhat3 is the prediction for this datetime, predicted 3 steps ago, "3 steps old".
    """
    # Receives df with single ID column
    assert len(df["ID"].unique()) == 1
    cols = ["ds", "y", "ID"]  # cols to keep from df
    df_forecast = pd.concat((df[cols],), axis=1)
    # create a line for each forecast_lag
    # 'yhat<i>' is the forecast for 'y' at 'ds' from i steps ago.
    for j in range(len(quantiles)):
        for forecast_lag in range(1, n_forecasts + 1):
            forecast = predicted[:, forecast_lag - 1, j]
            pad_before = max_lags + forecast_lag - 1
            pad_after = n_forecasts - forecast_lag
            yhat = np.pad(forecast, (pad_before, pad_after), mode="constant", constant_values=np.NaN)
            if prediction_frequency is not None:
                ds = df_forecast["ds"].iloc[pad_before : -pad_after if pad_after > 0 else None]
                mask = df_utils.create_mask_for_prediction_frequency(
                    prediction_frequency=prediction_frequency,
                    ds=ds,
                    forecast_lag=forecast_lag,
                )
                yhat = np.full((len(ds),), np.nan)
                yhat[mask] = forecast
                yhat = np.pad(yhat, (pad_before, pad_after), mode="constant", constant_values=np.NaN)
            # 0 is the median quantile index
            if j == 0:
                name = f"yhat{forecast_lag}"
            else:
                name = f"yhat{forecast_lag} {round(quantiles[j] * 100, 1)}%"
            df_forecast[name] = yhat

    if components is None:
        return df_forecast

    # else add components
    lagged_components = [
        "ar",
    ]
    if config_lagged_regressors is not None:
        for name in config_lagged_regressors.keys():
            lagged_components.append(f"lagged_regressor_{name}")
    for comp in lagged_components:
        if comp in components:
            for j in range(len(quantiles)):
                for forecast_lag in range(1, n_forecasts + 1):
                    forecast = components[comp][:, forecast_lag - 1, j]  # 0 is the median quantile
                    pad_before = max_lags + forecast_lag - 1
                    pad_after = n_forecasts - forecast_lag
                    yhat = np.pad(forecast, (pad_before, pad_after), mode="constant", constant_values=np.NaN)
                    if prediction_frequency is not None:
                        ds = df_forecast["ds"].iloc[pad_before : -pad_after if pad_after > 0 else None]
                        mask = df_utils.create_mask_for_prediction_frequency(
                            prediction_frequency=prediction_frequency,
                            ds=ds,
                            forecast_lag=forecast_lag,
                        )
                        yhat = np.full((len(ds),), np.nan)
                        yhat[mask] = forecast
                        yhat = np.pad(yhat, (pad_before, pad_after), mode="constant", constant_values=np.NaN)
                    if j == 0:  # temporary condition to add only the median component
                        name = f"{comp}{forecast_lag}"
                        df_forecast[name] = yhat

    # only for non-lagged components
    for comp in components:
        if comp not in lagged_components:
            for j in range(len(quantiles)):
                forecast_0 = components[comp][0, :, j]
                forecast_rest = components[comp][1:, n_forecasts - 1, j]
                yhat = np.pad(
                    np.concatenate((forecast_0, forecast_rest)), (max_lags, 0), mode="constant", constant_values=np.NaN
                )
                if prediction_frequency is not None:
                    date_list = []
                    for key, value in prediction_frequency.items():
                        if key == "daily-hour":
                            dates_comp = dates[dates.dt.hour == value]
                        elif key == "weekly-day":
                            dates_comp = dates[dates.dt.dayofweek == value]
                        elif key == "monthly-day":
                            dates_comp = dates[dates.dt.day == value]
                        elif key == "yearly-month":
                            dates_comp = dates[dates.dt.month == value]
                        elif key == "hourly-minute":
                            dates_comp = dates[dates.dt.minute == value]
                        else:
                            raise ValueError(f"prediction_frequency {key} not supported")
                        date_list.append(dates_comp)
                    # create new pd.Series only containing the dates that are in all Series in date_list
                    dates_comp = pd.Series(date_list[0])
                    for i in range(1, len(date_list)):
                        dates_comp = dates_comp[dates_comp.isin(date_list[i])]
                    ser = pd.Series(dtype="datetime64[ns]")
                    for date in dates_comp:
                        d = pd.date_range(date, periods=n_forecasts + 1, freq=freq)
                        ser = pd.concat((ser, pd.Series(d).iloc[1:]))
                    df_comp = pd.DataFrame({"ds": ser, "yhat": components[comp][:, :, j].flatten()}).drop_duplicates(
                        subset="ds"
                    )
                    df_comp, _ = df_utils.add_missing_dates_nan(df=df_comp, freq=freq)
                    yhat = pd.merge(df_forecast.filter(["ds", "ID"]), df_comp, on="ds", how="left")["yhat"].values
                if j == 0:  # temporary condition to add only the median component
                    # add yhat into dataframe, using df_forecast indexing
                    yhat_df = pd.Series(yhat, name=comp).set_axis(df_forecast.index)
                    df_forecast = pd.concat([df_forecast, yhat_df], axis=1, ignore_index=False)
    return df_forecast


def _convert_raw_predictions_to_raw_df(
    dates: pd.Series,
    predicted: np.ndarray,
    n_forecasts: int,
    quantiles: List[float],
    components: Optional[Components] = None,
) -> pd.DataFrame:
    """Turns forecast-origin-wise predictions into forecast-target-wise predictions.

    Parameters
    ----------
        dates : pd.Series
            timestamps referring to the start of the predictions.
        predicted : np.array
            Array containing the forecasts
        n_forecasts : int
            optional, number of steps ahead of prediction time step to forecast
        quantiles : list[float]
            optional, list of quantiles for quantile regression uncertainty estimate
        components : dict[np.array]
            Dictionary of components containing an array of each components' contribution to the forecast

    Returns
    -------
        pd. DataFrame
            columns ``ds``, ``y``, and [``step<i>``]

            Note
            ----
            where step<i> refers to the i-step-ahead prediction *made at* this row's datetime.
            e.g. the first forecast step0 is the prediction for this timestamp,
            the step1 is for the timestamp after, ...
            ... step3 is the prediction for 3 steps into the future,
            predicted using information up to (excluding) this datetime.
    """
    all_data = predicted
    df_raw = pd.DataFrame()
    df_raw.insert(0, "ds", dates.values)
    df_raw.insert(1, "ID", "__df__")  # type: ignore
    for forecast_lag in range(n_forecasts):
        for quantile_idx in range(len(quantiles)):
            # 0 is the median quantile index
            if quantile_idx == 0:
                step_name = f"step{forecast_lag}"
            else:
                step_name = f"step{forecast_lag} {quantiles[quantile_idx] * 100}%"
            data = all_data[:, forecast_lag, quantile_idx]
            ser = pd.Series(data=data, name=step_name)
            df_raw = df_raw.merge(ser, left_index=True, right_index=True)
        if components is not None:
            for comp_name, comp_data in components.items():
                comp_name_ = f"{comp_name}{forecast_lag}"
                data = comp_data[:, forecast_lag, 0]  # for components the quantiles are ignored for now
                ser = pd.Series(data=data, name=comp_name_)
                df_raw = df_raw.merge(ser, left_index=True, right_index=True)
    return df_raw


def _prepare_dataframe_to_predict(model, df: pd.DataFrame, max_lags: int, freq: Optional[str]) -> pd.DataFrame:
    """
    Pre-processes a dataframe for prediction using the specified model.

    Parameters
    ----------
        model:
            The NeuralProphet model
        df: pd.DataFrame
            dataframe containing column ``ds``, ``y``, and optionally``ID`` with all data
        max_lags: int
            The maximum number of lags to include in the output dataframe.
        freq: str
            data step sizes. Frequency of data recording,

    Returns
    ----------
        pd.DataFrame
            pre-processed dataframe

    Raises
    ----------
        ValueError
        If the input dataframe has already been normalized, if there is insufficient input data for prediction,
        if only datestamps are provided but y values are needed for auto-regression.
    """
    # Receives df with ID column
    df_prepared = pd.DataFrame()
    for df_name, df_i in df.groupby("ID"):
        df_i = df_i.copy(deep=True)
        _ = df_utils.infer_frequency(df_i, n_lags=max_lags, freq=freq)
        # check if received pre-processed df
        if "y_scaled" in df_i.columns or "t" in df_i.columns:
            raise ValueError(
                "DataFrame has already been normalized. " "Please provide raw dataframe or future dataframe."
            )
        # Checks
        if len(df_i) == 0 or len(df_i) < max_lags:
            raise ValueError(
                "Insufficient input data for a prediction."
                "Please supply historic observations (number of rows) of at least max_lags (max of number of n_lags)."
            )
        if len(df_i.columns) == 1 and "ds" in df_i:
            if max_lags != 0:
                raise ValueError("only datestamps provided but y values needed for auto-regression.")
            df_i = _check_dataframe(model, df_i, check_y=False, exogenous=False)
        else:
            df_i = _check_dataframe(model, df_i, check_y=model.max_lags > 0, exogenous=False)
            # fill in missing nans except for nans at end
            df_i = _handle_missing_data(
                df=df_i,
                freq=freq,
                n_lags=model.n_lags,
                n_forecasts=model.n_forecasts,
                config_missing=model.config_missing,
                config_regressors=model.config_regressors,
                config_lagged_regressors=model.config_lagged_regressors,
                config_events=model.config_events,
                config_seasonality=model.config_seasonality,
                predicting=True,
            )
        df_prepared = pd.concat((df_prepared, df_i.copy(deep=True).reset_index(drop=True)), ignore_index=True)
    return df_prepared


def _validate_column_name(
    name: str,
    config_events: Optional[ConfigEvents],
    config_country_holidays: Optional[ConfigCountryHolidays],
    config_seasonality: Optional[ConfigSeasonality],
    config_lagged_regressors: Optional[ConfigLaggedRegressors],
    config_regressors: Optional[ConfigFutureRegressors],
    events: Optional[bool] = True,
    seasons: Optional[bool] = True,
    regressors: Optional[bool] = True,
    covariates: Optional[bool] = True,
):
    """Validates the name of a seasonality, event, or regressor.

    Parameters
    ----------
        name : str
            name of seasonality, event or regressor
        config_events : Optional[ConfigEvents]
            Configuration options for adding events to the model.
        config_country_holidays : Optional[ConfigCountryHolidays]
            Configuration options for adding country holidays to the model.
        config_seasonality : Optional[ConfigSeasonality]
            Configuration options for adding seasonal components to the model.
        config_lagged_regressors : Optional[ConfigLaggedRegressors]
            Configuration options for adding lagged external regressors to the model.
        config_regressors : Optional[ConfigFutureRegressors]
            Configuration options for adding future regressors to the model.
        events : bool
            check if name already used for event
        seasons : bool
            check if name already used for seasonality
        regressors : bool
            check if name already used for regressor
        covariates : bool
            check if name already used for covariate
    """
    reserved_names = [
        "trend",
        "additive_terms",
        "daily",
        "weekly",
        "yearly",
        "events",
        "holidays",
        "zeros",
        "extra_regressors_additive",
        "yhat",
        "extra_regressors_multiplicative",
        "multiplicative_terms",
        "ID",
    ]
    rn_l = [n + "_lower" for n in reserved_names]
    rn_u = [n + "_upper" for n in reserved_names]
    reserved_names.extend(rn_l)
    reserved_names.extend(rn_u)
    reserved_names.extend(["ds", "y", "cap", "floor", "y_scaled", "cap_scaled"])
    if name in reserved_names:
        raise ValueError(f"Name {name!r} is reserved.")
    if events and config_events is not None:
        if name in config_events.keys():
            raise ValueError(f"Name {name!r} already used for an event.")
    if events and config_country_holidays is not None:
        if name in config_country_holidays.holiday_names:
            raise ValueError(f"Name {name!r} is a holiday name in {config_country_holidays.country}.")
    if seasons and config_seasonality is not None:
        if name in config_seasonality.periods:
            raise ValueError(f"Name {name!r} already used for a seasonality.")
    if covariates and config_lagged_regressors is not None:
        if name in config_lagged_regressors:
            raise ValueError(f"Name {name!r} already used for an added covariate.")
    if regressors and config_regressors.regressors is not None:
        if name in config_regressors.regressors.keys():
            raise ValueError(f"Name {name!r} already used for an added regressor.")


def _check_dataframe(
    model,
    df: pd.DataFrame,
    check_y: bool = True,
    exogenous: bool = True,
    future: Optional[bool] = None,
) -> pd.DataFrame:
    """Performs basic data sanity checks and ordering

    Prepare dataframe for fitting or predicting.

    Parameters
    ----------
        df : pd.DataFrame
            dataframe containing column ``ds``, ``y``, and optionally``ID`` with all data
        check_y : bool
            if df must have series values

            Note
            ----
            set to True if training or predicting with autoregression
        exogenous : bool
            whether to check covariates, regressors and events column names
        future : bool
            whether this function is called by make_future_dataframe()

    Returns
    -------
        pd.DataFrame
            checked dataframe
    """
    if len(df) < (model.n_forecasts + model.n_lags) and not future:
        raise ValueError(
            "Dataframe has less than n_forecasts + n_lags rows. "
            "Forecasting not possible. Please either use a larger dataset, or adjust the model parameters."
        )
    df, _, _, _ = df_utils.prep_or_copy_df(df)
    df, regressors_to_remove, lag_regressors_to_remove = df_utils.check_dataframe(
        df=df,
        check_y=check_y,
        covariates=model.config_lagged_regressors if exogenous else None,
        regressors=model.config_regressors.regressors if exogenous else None,
        events=model.config_events if exogenous else None,
        seasonalities=model.config_seasonality if exogenous else None,
        future=True if future else None,
    )

    if model.config_regressors.regressors is not None:
        for reg in regressors_to_remove:
            log.warning(f"Removing regressor {reg} because it is not present in the data.")
            model.config_regressors.regressors.pop(reg)
        if model.config_regressors.regressors is not None and len(model.config_regressors.regressors) == 0:
            model.config_regressors.regressors = None
    if model.config_lagged_regressors is not None:
        for reg in lag_regressors_to_remove:
            log.warning(f"Removing lagged regressor {reg} because it is not present in the data.")
            model.config_lagged_regressors.pop(reg)
        if len(model.config_lagged_regressors) == 0:
            model.config_lagged_regressors = None
    return df


def _handle_missing_data(
    df: pd.DataFrame,
    freq: Optional[str],
    n_lags: int,
    n_forecasts: int,
    config_missing,
    config_regressors: Optional[ConfigFutureRegressors],
    config_lagged_regressors: Optional[ConfigLaggedRegressors],
    config_events: Optional[ConfigEvents],
    config_seasonality: Optional[ConfigSeasonality],
    predicting: bool = False,
) -> pd.DataFrame:
    """
    Checks and normalizes new data, auto-imputing missing data if `config_missing` allows it.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe containing columns 'ds', 'y', and optionally 'ID' with all data.
    freq : str
            data step sizes. Frequency of data recording,

            Note
            ----
            Any valid frequency for pd.date_range, such as ``5min``, ``D``, ``MS`` or ``auto`` (default) to
            automatically set frequency.
    n_lags : int
        Previous time series steps to include in auto-regression. Aka AR-order
    n_forecasts : int
        Number of steps ahead of prediction time step to forecast.
    config_missing :
        Configuration options for handling missing data.
    config_regressors : Optional[ConfigFutureRegressors]
        Configuration options for adding future regressors to the model.
    config_lagged_regressors : Optional[ConfigLaggedRegressors]
        Configuration options for adding lagged external regressors to the model.
    config_events : Optional[ConfigEvents]
        Configuration options for adding events to the model.
    config_seasonality : Optional[ConfigSeasonality]
        Configuration options for adding seasonal components to the model.
    predicting : bool, default False
        If True, allows missing values in the 'y' column for the forecast period, or missing completely.

    Returns
    -------
    pd.DataFrame
        The pre-processed DataFrame, including imputed missing data, if applicable.

    """
    df, _, _, _ = df_utils.prep_or_copy_df(df)

    if n_lags == 0 and not predicting:
        # drop rows with NaNs in y and count them
        df_na_dropped = df.dropna(subset=["y"])
        n_dropped = len(df) - len(df_na_dropped)
        if n_dropped > 0:
            df = df_na_dropped
            log.info(f"Dropped {n_dropped} rows with NaNs in 'y' column.")

    if n_lags > 0:
        # add missig dates to df
        df_grouped = df.groupby("ID").apply(lambda x: x.set_index("ds").resample(freq).asfreq()).drop(columns=["ID"])
        n_missing_dates = len(df_grouped) - len(df)
        if n_missing_dates > 0:
            df = df_grouped.reset_index()
            log.info(f"Added {n_missing_dates} missing dates.")

    if config_regressors is not None and config_regressors.regressors is not None:
        # drop complete row for future regressors that are NaN at the end
        last_valid_index = df.groupby("ID")[list(config_regressors.regressors.keys())].apply(
            lambda x: x.last_valid_index()
        )
        df_dropped = df.groupby("ID", group_keys=False).apply(lambda x: x.loc[: last_valid_index[x.name]])
        n_dropped = len(df) - len(df_dropped)
        if n_dropped > 0:
            df = df_dropped
            log.info(f"Dropped {n_dropped} rows at the end with NaNs in future regressors.")

    dropped_trailing_y = False
    if df["y"].isna().any():
        # drop complete row if y of ID ends with nan
        last_valid_index = df.groupby("ID")["y"].apply(lambda x: x.last_valid_index())
        df_dropped = df.groupby("ID", group_keys=False).apply(lambda x: x.loc[: last_valid_index[x.name]])
        n_dropped = len(df) - len(df_dropped)
        if n_dropped > 0:
            dropped_trailing_y = True
            # save dropped rows for later
            df_to_add = df.groupby("ID", group_keys=False).apply(lambda x: x.loc[last_valid_index[x.name] + 1 :])
            df = df_dropped
            log.info(f"Dropped {n_dropped} rows at the end with NaNs in 'y' column.")

    if config_missing.impute_missing:
        # impute missing values
        data_columns = []
        if n_lags > 0:
            data_columns.append("y")
        if config_lagged_regressors is not None:
            data_columns.extend(config_lagged_regressors.keys())
        if config_regressors is not None and config_regressors.regressors is not None:
            data_columns.extend(config_regressors.regressors.keys())
        if config_events is not None:
            data_columns.extend(config_events.keys())
        conditional_cols = []
        if config_seasonality is not None:
            conditional_cols = list(
                set(
                    [
                        value.condition_name
                        for key, value in config_seasonality.periods.items()
                        if value.condition_name is not None
                    ]
                )
            )
            data_columns.extend(conditional_cols)
        for column in data_columns:
            sum_na = df[column].isna().sum()
            if sum_na > 0:
                log.warning(f"{sum_na} missing values in column {column} were detected in total. ")
                # use 0 substitution for holidays and events missing values
                if config_events is not None and column in config_events.keys():
                    df[column].fillna(0, inplace=True)
                    remaining_na = 0
                else:
                    df.loc[:, column], remaining_na = df_utils.fill_linear_then_rolling_avg(
                        df[column],
                        limit_linear=config_missing.impute_linear,
                        rolling=config_missing.impute_rolling,
                    )
                log.info(f"{sum_na - remaining_na} NaN values in column {column} were auto-imputed.")
                if remaining_na > 0:
                    log.warning(
                        f"More than {2 * config_missing.impute_linear + config_missing.impute_rolling} consecutive \
                            missing values encountered in column {column}. "
                        f"{remaining_na} NA remain after auto-imputation. "
                    )
    if dropped_trailing_y and predicting:
        # add trailing y values again if in predict mode
        df = pd.concat([df, df_to_add])
        if config_seasonality is not None and len(conditional_cols) > 0:
            df[conditional_cols] = df[conditional_cols].ffill()  # type: ignore
    return df


def _create_dataset(model, df, predict_mode, prediction_frequency=None):
    """Construct dataset from dataframe.

    (Configured Hyperparameters can be overridden by explicitly supplying them.
    Useful to predict a single model component.)

    Parameters
    ----------
        df : pd.DataFrame
            dataframe containing column ``ds``, ``y``, and optionally``ID`` and
            normalized columns normalized columns ``ds``, ``y``, ``t``, ``y_scaled``
        predict_mode : bool
            specifies predict mode

            Options
                * ``False``: includes target values.
                * ``True``: does not include targets but includes entire dataset as input

        prediction_frequency: dict
            periodic interval in which forecasts should be made.
            Key: str
                periodicity of the predictions to be made, e.g. 'daily-hour'.

            Options
                * ``'hourly-minute'``: forecast once per hour at a specified minute
                * ``'daily-hour'``: forecast once per day at a specified hour
                * ``'weekly-day'``: forecast once per week at a specified day
                * ``'monthly-day'``: forecast once per month at a specified day
                * ``'yearly-month'``: forecast once per year at a specified month

            value: int
                forecast origin of the predictions to be made, e.g. 7 for 7am in case of 'daily-hour'.

    Returns
    -------
        TimeDataset
    """
    df, _, _, _ = df_utils.prep_or_copy_df(df)
    return time_dataset.GlobalTimeDataset(
        df,
        predict_mode=predict_mode,
        n_lags=model.n_lags,
        n_forecasts=model.n_forecasts,
        predict_steps=model.predict_steps,
        config_seasonality=model.config_seasonality,
        config_events=model.config_events,
        config_country_holidays=model.config_country_holidays,
        config_lagged_regressors=model.config_lagged_regressors,
        config_regressors=model.config_regressors,
        config_missing=model.config_missing,
        prediction_frequency=prediction_frequency,
    )



import torch
from torch.nn.modules.loss import _Loss


class PinballLoss(_Loss):
    """Class for the PinBall loss for quantile regression"""

    def __init__(self, loss_func, quantiles):
        """
        Args:
            loss_func : torch.nn._Loss
                Loss function to be used as the
                base loss for pinball loss
            quantiles : list
                list of quantiles estimated from the model
        """
        super().__init__()
        self.loss_func = loss_func
        self.quantiles = quantiles

    def forward(self, outputs, target):
        """
        Computes the pinball loss from forecasts
        Args:
            outputs : torch.tensor
                outputs from the model of dims (batch, no_quantiles, n_forecasts)
            target : torch.tensor
                actual targets of dims (batch, n_forecasts)

        Returns:
            float
                pinball loss
        """
        target = target.repeat(1, 1, len(self.quantiles))  # increase the quantile dimension of the targets
        differences = target - outputs
        base_losses = self.loss_func(outputs, target)  # dimensions - [n_batch, n_forecasts, no. of quantiles]
        positive_losses = (
            torch.tensor(self.quantiles, device=target.device).unsqueeze(dim=0).unsqueeze(dim=0) * base_losses
        )
        negative_losses = (
            1 - torch.tensor(self.quantiles, device=target.device).unsqueeze(dim=0).unsqueeze(dim=0)
        ) * base_losses
        pinball_losses = torch.where(differences >= 0, positive_losses, negative_losses)
        multiplier = torch.ones(size=(1, 1, len(self.quantiles)), device=target.device)
        multiplier[:, :, 0] = 2
        pinball_losses = multiplier * pinball_losses  # double the loss for the median quantile
        return pinball_losses



import logging
import warnings
from collections import OrderedDict
from typing import Optional

import numpy as np
import torch

from neuralprophet import time_dataset, utils_torch

log = logging.getLogger("NP.plotting")


def log_warning_deprecation_plotly(plotting_backend):
    if plotting_backend == "matplotlib":
        log.warning(
            "DeprecationWarning: default plotting_backend will be changed to plotly in a future version. "
            "Switch to plotly by calling `m.set_plotting_backend('plotly')`."
        )


def log_warning_resampler_invalid_env():
    log.warning(
        "Warning: plotly-resampler not supported for the environment you are using. "
        "Consider switching plotting_backend to 'plotly' or 'matplotlib "
    )


def log_warning_resampler_switch_to_valid_env():
    log.warning(
        "Warning: plotly-resampler not supported for the environment you are using. "
        "Plotting backend automatically switched to 'plotly' without resampling "
    )


def set_y_as_percent(ax):
    """Set y axis as percentage

    Parameters
    ----------
        ax : matplotlib axis
            Respective y axis element

    Returns
    -------
        matplotlib axis
            Manipulated axis element
    """
    warnings.filterwarnings(
        action="ignore", category=UserWarning
    )  # workaround until there is clear direction how to handle this recent matplotlib bug
    yticks = 100 * ax.get_yticks()
    yticklabels = [f"{y:.4g}%" for y in yticks]
    ax.set_yticklabels(yticklabels)
    return ax


def predict_one_season(m, quantile, name, n_steps=100, df_name="__df__"):
    """
     Predicts the seasonal component given a number of time steps.

    Parameters
    ----------
         m : NeuralProphet
             Fitted NeuralProphet model
         quantile: float
             The quantile for which the season is predicted
         name: str
             Name of seasonality component
         n_steps: int
             number of prediction steps related to the season frequency
         df_name: str
                Name of dataframe to refer to data params from original keys of train dataframes

        Returns
        -------
            t_i: np.array
                 time scale of predicted seasonal component
            predicted: OrderedDict
                 predicted seasonal component

    """
    config = m.config_seasonality.periods[name]
    t_i = np.arange(n_steps + 1) / float(n_steps)
    features = time_dataset.fourier_series_t(
        t=t_i * config.period, period=config.period, series_order=config.resolution
    )
    features = torch.from_numpy(np.expand_dims(features, 1))

    if df_name == "__df__":
        meta_name_tensor = None
    else:
        meta = OrderedDict()
        meta["df_name"] = [df_name for _ in range(n_steps + 1)]
        meta_name_tensor = torch.tensor([m.model.id_dict[i] for i in meta["df_name"]])

    quantile_index = m.model.quantiles.index(quantile)
    predicted = m.model.seasonality.compute_fourier(features=features, name=name, meta=meta_name_tensor)[
        :, :, quantile_index
    ]
    predicted = predicted.squeeze().detach().numpy()
    if m.config_seasonality.mode == "additive":
        data_params = m.config_normalization.get_data_params(df_name)
        scale = data_params["y"].scale
        predicted = predicted * scale
    return t_i, predicted


def predict_season_from_dates(m, dates, name, quantile, df_name="__df__"):
    """
     Predicts the seasonal component given a date range.

     Parameters
     ----------
         m : NeuralProphet
             Fitted NeuralProphet model
         dates: pd.datetime
             date range for prediction
         name: str
             Name of seasonality component
         quantile: float
             The quantile for which the season is predicted
         df_name: str
             Name of dataframe to refer to data params from original keys of train dataframes

    Returns
    -------
        predicted: OrderedDict
             presdicted seasonal component
    """
    config = m.config_seasonality.periods[name]
    features = time_dataset.fourier_series(dates=dates, period=config.period, series_order=config.resolution)
    features = torch.from_numpy(np.expand_dims(features, 1))
    if df_name == "__df__":
        meta_name_tensor = None
    else:
        meta = OrderedDict()
        meta["df_name"] = [df_name for _ in range(len(dates))]
        meta_name_tensor = torch.tensor([m.model.id_dict[i] for i in meta["df_name"]])

    quantile_index = m.model.quantiles.index(quantile)
    predicted = m.model.seasonality.compute_fourier(features=features, name=name, meta=meta_name_tensor)[
        :, :, quantile_index
    ]

    predicted = predicted.squeeze().detach().numpy()
    if m.config_seasonality.mode == "additive":
        data_params = m.config_normalization.get_data_params(df_name)
        scale = data_params["y"].scale
        predicted = predicted * scale
    predicted = {name: predicted}
    return predicted


def check_if_configured(m, components, error_flag=False):  # move to utils
    """Check if components were set in the model configuration by the user.

    Parameters
    ----------
        m : NeuralProphet
            Fitted NeuralProphet model
        components : str or list, optional
            name or list of names of components to check

            Options
            ----
            * ``trend``
            * ``trend_rate_change``
            * ``seasonality``
            * ``autoregression``
            * ``lagged_regressors```
            * ``events``
            * ``future_regressors`
            * ``uncertainty``
        error_flag : bool
            Activate to raise a ValueError if component has not been configured

    Returns
    -------
        components
            list of components only including the components set in the model configuration
    """
    invalid_components = []
    if "trend_rate_change" in components and m.model.config_trend.changepoints is None:
        components.remove("trend_rate_change")
        invalid_components.append("trend_rate_change")
    if "seasonality" in components and m.config_seasonality is None:
        components.remove("seasonality")
        invalid_components.append("seasonality")
    if "autoregression" in components and not m.config_ar.n_lags > 0:
        components.remove("autoregression")
        invalid_components.append("autoregression")
    if "lagged_regressors" in components and m.config_lagged_regressors is None:
        components.remove("lagged_regressors")
        invalid_components.append("lagged_regressors")
    if "events" in components and (m.config_events is None and m.config_country_holidays is None):
        components.remove("events")
        invalid_components.append("events")
    if "future_regressors" in components and m.config_regressors.regressors is None:
        components.remove("future_regressors")
        invalid_components.append("future_regressors")
    if "uncertainty" in components and not len(m.model.quantiles) > 1:
        components.remove("uncertainty")
        invalid_components.append("uncertainty")
    if error_flag and len(invalid_components) != 0:
        raise ValueError(
            f" Selected component(s) {(invalid_components)} for plotting not specified in the model configuration."
        )
    return components


def get_valid_configuration(  # move to utils
    m, components=None, df_name=None, valid_set=None, validator=None, forecast_in_focus=None, quantile=0.5
):
    """Validate and adapt the selected components to be plotted.

    Parameters
    ----------
        m : NeuralProphet
            Fitted NeuralProphet model
        components : str or list, optional
            name or list of names of components to validate and adapt
        df_name: str
            ID from time series that should be plotted
        valid_set : str or list, optional
            name or list of names of components that are defined as valid option

            Options
            ----
             * (default)``None``:  All components the user set in the model configuration are validated and adapted
            * ``trend``
            * ``seasonality``
            * ``autoregression``
            * ``lagged_regressors``
            * ``future_regressors``
            * ``events``
            * ``uncertainty``
        validator: str
            specifies the validation purpose to customize

            Options
            ----
            * ``plot_parameters``: customize for plot_parameters() function
            * ``plot_components``: customize for plot_components() function
        forecast_in_focus: int
            optinal, i-th step ahead forecast to plot

            Note
            ----
            None (default): plot self.highlight_forecast_step_n by default
        quantile: float
            The quantile for which the model parameters are to be plotted

            Note
            ----
            0.5 (default):  Parameters will be plotted for the median quantile.

    Returns
    -------
        valid_configuration: dict
            dict of validated components and values to be plotted
    """
    if not isinstance(valid_set, list):
        valid_set = [valid_set]

    if components is None:
        components = valid_set
        components = check_if_configured(m=m, components=components)
    else:
        if not isinstance(components, list):
            components = [components]
        components = [comp.lower() for comp in components]
        for comp in components:
            if comp not in valid_set:
                raise ValueError(
                    f" Selected component {comp} is either mis-spelled or not an available "
                    f"option for this function."
                )
        components = check_if_configured(m=m, components=components, error_flag=True)
    if validator is None:
        raise ValueError("Specify a validator from the available options")
    # Adapt Normalization
    if validator == "plot_parameters":
        # Set to True in case of local normalization and unknown_data_params is not True
        overwriting_unknown_data_normalization = False
        if m.config_normalization.global_normalization:
            if df_name is None and m.id_list.__len__() == 1:
                df_name = "__df__"
            elif df_name is None and m.id_list.__len__() > 1:
                df_name = m.id_list[0]
            else:
                log.debug("Global normalization set - ignoring given df_name for normalization")
        else:
            if df_name is None:
                if m.id_list.__len__() > 1:
                    if (
                        m.model.config_seasonality.global_local in ["local", "glocal"]
                        or m.model.config_trend.trend_global_local == "local"
                    ):
                        df_name = m.id_list
                        log.warning(
                            "Glocal model set with > 1 time series in the pd.DataFrame. Plotting components of mean \
                                time series and quants. "
                        )
                    else:
                        df_name = m.id_list[0]
                else:
                    log.warning("Local normalization set, but df_name is None. Using global data params instead.")
                    df_name = "__df__"
                if not m.config_normalization.unknown_data_normalization:
                    m.config_normalization.unknown_data_normalization = True
                    overwriting_unknown_data_normalization = True
            elif df_name not in m.config_normalization.local_data_params:
                log.warning(
                    f"Local normalization set, but df_name '{df_name}' not found. Using global data params instead."
                )
                df_name = "__df__"
                if not m.config_normalization.unknown_data_normalization:
                    m.config_normalization.unknown_data_normalization = True
                    overwriting_unknown_data_normalization = True
            else:
                log.debug(f"Local normalization set. Data params for {df_name} will be used to denormalize.")

    # Identify components to be plotted
    # as dict, minimum: {plot_name}
    plot_components = []
    if validator == "plot_parameters":
        quantile_index = m.model.quantiles.index(quantile)

    # Plot trend
    if "trend" in components:
        plot_components.append({"plot_name": "Trend", "comp_name": "trend"})
    if "trend_rate_change" in components:
        plot_components.append({"plot_name": "Trend Rate Change"})

    # Plot  seasonalities, if present
    if "seasonality" in components:
        for name in m.config_seasonality.periods:
            if validator == "plot_components":
                plot_components.append(
                    {
                        "plot_name": f"{name} seasonality",
                        "comp_name": name,
                    }
                )
            elif validator == "plot_parameters":
                plot_components.append({"plot_name": "seasonality", "comp_name": name})

    # AR
    if "autoregression" in components:
        if validator == "plot_components":
            if forecast_in_focus is None:
                plot_components.append(
                    {
                        "plot_name": "Auto-Regression",
                        "comp_name": "ar",
                        "num_overplot": m.n_forecasts,
                        "bar": True,
                    }
                )
            else:
                plot_components.append(
                    {
                        "plot_name": f"AR ({forecast_in_focus})-ahead",
                        "comp_name": f"ar{forecast_in_focus}",
                    }
                )
        elif validator == "plot_parameters":
            plot_components.append(
                {
                    "plot_name": "lagged weights",
                    "comp_name": "AR",
                    "weights": utils_torch.interprete_model(m.model, net="ar_net", forward_func="auto_regression")
                    .detach()
                    .numpy(),
                    "focus": forecast_in_focus,
                }
            )

    # Add lagged regressors
    lagged_scalar_regressors = []
    if "lagged_regressors" in components:
        if validator == "plot_components":
            if forecast_in_focus is None:
                for name in m.config_lagged_regressors.keys():
                    plot_components.append(
                        {
                            "plot_name": f'Lagged Regressor "{name}"',
                            "comp_name": f"lagged_regressor_{name}",
                            "num_overplot": m.n_forecasts,
                            "bar": True,
                        }
                    )
            else:
                for name in m.config_lagged_regressors.keys():
                    plot_components.append(
                        {
                            "plot_name": f'Lagged Regressor "{name}" ({forecast_in_focus})-ahead',
                            "comp_name": f"lagged_regressor_{name}{forecast_in_focus}",
                        }
                    )
        elif validator == "plot_parameters":
            for name in m.config_lagged_regressors.keys():
                if m.config_lagged_regressors[name].as_scalar:
                    lagged_scalar_regressors.append((name, m.model.get_covar_weights()[name].detach().numpy()))
                else:
                    plot_components.append(
                        {
                            "plot_name": "lagged weights",
                            "comp_name": f'Lagged Regressor "{name}"',
                            "weights": m.model.get_covar_weights()[name].detach().numpy(),
                            "focus": forecast_in_focus,
                        }
                    )

    # Add Events
    additive_events = []
    multiplicative_events = []
    if "events" in components:
        additive_events_flag = False
        muliplicative_events_flag = False
        if m.config_events is not None:
            for event, configs in m.config_events.items():
                if validator == "plot_components" and configs.mode == "additive":
                    additive_events_flag = True
                elif validator == "plot_components" and configs.mode == "multiplicative":
                    muliplicative_events_flag = True
                elif validator == "plot_parameters":
                    event_params = m.model.get_event_weights(event)
                    weight_list = [
                        (key, param.detach().numpy()[quantile_index, :]) for key, param in event_params.items()
                    ]
                    if configs.mode == "additive":
                        additive_events = additive_events + weight_list
                    elif configs.mode == "multiplicative":
                        multiplicative_events = multiplicative_events + weight_list

        if m.config_country_holidays is not None:
            for country_holiday in m.config_country_holidays.holiday_names:
                if validator == "plot_components" and m.config_country_holidays.mode == "additive":
                    additive_events_flag = True
                elif validator == "plot_components" and m.config_country_holidays.mode == "multiplicative":
                    muliplicative_events_flag = True
                elif validator == "plot_parameters":
                    event_params = m.model.get_event_weights(country_holiday)
                    weight_list = [
                        (key, param.detach().numpy()[quantile_index, :]) for key, param in event_params.items()
                    ]
                    if m.config_country_holidays.mode == "additive":
                        additive_events = additive_events + weight_list
                    elif m.config_country_holidays.mode == "multiplicative":
                        multiplicative_events = multiplicative_events + weight_list

        if additive_events_flag:
            plot_components.append(
                {
                    "plot_name": "Additive Events",
                    "comp_name": "events_additive",
                }
            )
        if muliplicative_events_flag:
            plot_components.append(
                {
                    "plot_name": "Multiplicative Events",
                    "comp_name": "events_multiplicative",
                    "multiplicative": True,
                }
            )

    # Add Regressors
    additive_future_regressors = []
    multiplicative_future_regressors = []
    if "future_regressors" in components:
        for regressor, configs in m.config_regressors.regressors.items():
            if validator == "plot_components" and configs.mode == "additive":
                plot_components.append(
                    {
                        "plot_name": "Additive Future Regressors",
                        "comp_name": "future_regressors_additive",
                    }
                )
            elif validator == "plot_components" and configs.mode == "multiplicative":
                plot_components.append(
                    {
                        "plot_name": "Multiplicative Future Regressors",
                        "comp_name": "future_regressors_multiplicative",
                        "multiplicative": True,
                    }
                )
            elif validator == "plot_parameters":
                regressor_param = m.model.future_regressors.get_reg_weights(regressor)[quantile_index, :]
                if configs.mode == "additive":
                    additive_future_regressors.append((regressor, regressor_param.detach().numpy()))
                elif configs.mode == "multiplicative":
                    multiplicative_future_regressors.append((regressor, regressor_param.detach().numpy()))

    # Plot  quantiles as a separate component, if present
    # If multiple steps in the future are predicted, only plot quantiles if highlight_forecast_step_n is set
    if (
        "quantiles" in components
        and validator == "plot_components"
        and len(m.model.quantiles) > 1
        and forecast_in_focus is None
    ):
        if len(m.config_train.quantiles) > 1 and (
            m.n_forecasts > 1 or m.config_ar.n_lags > 0
        ):  # rather query if n_forecasts >1 than n_lags>1
            raise ValueError(
                "Please specify step_number using the highlight_nth_step_ahead_of_each_forecast function"
                " for quantiles plotting when autoregression enabled."
            )
        for i in range(1, len(m.model.quantiles)):
            plot_components.append(
                {
                    "plot_name": "Uncertainty",
                    "comp_name": f"yhat1 {round(m.model.quantiles[i] * 100, 1)}%",
                    "fill": True,
                }
            )
    elif (
        "uncertainty" in components
        and validator == "plot_components"
        and len(m.model.quantiles) > 1
        and forecast_in_focus is not None
    ):
        for i in range(1, len(m.model.quantiles)):
            plot_components.append(
                {
                    "plot_name": "Uncertainty",
                    "comp_name": f"yhat{forecast_in_focus} {round(m.model.quantiles[i] * 100, 1)}%",
                    "fill": True,
                }
            )
    if validator == "plot_parameters":
        if len(additive_future_regressors) > 0:
            plot_components.append({"plot_name": "Additive future regressor"})
        if len(multiplicative_future_regressors) > 0:
            plot_components.append({"plot_name": "Multiplicative future regressor"})
        if len(lagged_scalar_regressors) > 0:
            plot_components.append({"plot_name": "Lagged scalar regressor"})
        if len(additive_events) > 0:
            data_params = m.config_normalization.get_data_params(df_name)
            scale = data_params["y"].scale
            additive_events = [(key, weight * scale) for (key, weight) in additive_events]
            plot_components.append({"plot_name": "Additive event"})
        if len(multiplicative_events) > 0:
            plot_components.append({"plot_name": "Multiplicative event"})

        valid_configuration = {
            "components_list": plot_components,
            "additive_future_regressors": additive_future_regressors,
            "additive_events": additive_events,
            "multiplicative_future_regressors": multiplicative_future_regressors,
            "multiplicative_events": multiplicative_events,
            "lagged_scalar_regressors": lagged_scalar_regressors,
            "overwriting_unknown_data_normalization": overwriting_unknown_data_normalization,
            "df_name": df_name,
        }
    elif validator == "plot_components":
        valid_configuration = {
            "components_list": plot_components,
        }
    return valid_configuration


def validate_current_env_for_resampler(auto: bool = False) -> Optional[bool]:
    """
    Validate the current environment to check if it is a valid environment for plotly-resampler and if invalid trigger
    warning message.

    Parameters
    ----------
    auto: bool, optional
        If True, the function will automatically switch to a valid environment if the current environment is not valid.
        If False, the function will return None if the current environment is not valid.
    Returns
    -------
    bool :
        True if the current environment is a valid environment to run the code, False if the current environment is
        not a valid environment to run the code. None if the current environment is not a valid environment to run
        the code and the function did not switch to a valid environment.
    """

    from IPython import get_ipython

    if "google.colab" in str(get_ipython()):
        if auto:
            log_warning_resampler_switch_to_valid_env()
            valid_env = False
        else:
            log_warning_resampler_invalid_env()
            valid_env = None
    else:
        if is_notebook():
            valid_env = True
        else:
            if auto:
                log_warning_resampler_switch_to_valid_env()
                valid_env = False
            else:
                log_warning_resampler_invalid_env()
                valid_env = None
    return valid_env


def is_notebook():
    """
    Determine if the code is being executed in a Jupyter notebook environment.

    Returns
    -------
    bool :
        True if the code is being executed in a Jupyter notebook, False otherwise.
    """
    try:
        from IPython import get_ipython

        if "IPKernelApp" not in get_ipython().config:  # pragma: no cover
            return False
    except ImportError:
        return False
    except AttributeError:
        return False
    return True


def select_plotting_backend(model, plotting_backend):
    """Automatically selects the plotting backend based on the global plotting_backend and plotting_backend set by the
    user. If the plotting backend is selected as "plotly-resampler", triggers warning message.

    Parameters
    ----------
    model: NeuralProphet
        The configured model.
    plotting_backend: str
        The plotting backend to use.

    Returns
    -------
    str
        The new plotting backend.
    """
    if hasattr(model, "plotting_backend") and plotting_backend is None:
        plotting_backend = model.plotting_backend
        if plotting_backend == "plotly-resampler":
            validate_current_env_for_resampler()
    else:
        if plotting_backend is None:
            if validate_current_env_for_resampler(auto=True):
                plotting_backend = "plotly-resampler"
            else:
                plotting_backend = "plotly"
        elif plotting_backend == "plotly-resampler":
            validate_current_env_for_resampler()
    return plotting_backend.lower()
